<?php

// example usage of gpollsurveyresearchtest_widget_override filter to allow widgets to override form settings
/*
add_filter('gpollsurveyresearchtest_widget_override' , 'gpollsurveyresearchtest_widget_override');

function gpollsurveyresearchtest_widget_override($form_ids){
     array_push($form_ids, 7, 9);
     return $form_ids;
}
*/


// The wp_cron job will take over the calculation of the results if 5 seconds is not long enough at the time of form submission/update.
// The default schedule is set to hourly to avoid heavy loads on the server but you can change this frequency by using the gform_polls_cron_schedule.
// The format follows the format required by the WordPress cron_schedules filter. http://codex.wordpress.org/Function_Reference/wp_get_schedules
// Important: the Polls Add-On must be deactivated and reactivated in order to reschedule the task.
/*
add_filter( 'gform_polls_cron_schedule', 'gform_polls_cron_add_twice_hourly' );
function gform_polls_cron_add_twice_hourly($schedules) {
    // Adds twice hourly to the existing schedules.
    $schedules['twicehourly'] = array(
        'interval' => 1800, // number of seconds in the interval
        'display' => 'Twice Hourly'
    );
    return $schedules;
}
*/

// By default the percentage will round to the nearest whole number
// This can be overridden with the gform_polls_percentage_precision hook

/*
add_filter('gform_polls_percentage_precision', 'gform_polls_custom_precision', 10, 2);
function gform_polls_custom_precision($precision, $form_id){
    return 1;
}
*/


//------------------------------------------

GFForms::include_addon_framework();

require_once( 'pollwidget.php' );

register_deactivation_hook( __FILE__, array( 'GFPollsSurveysResearchTests', 'remove_wp_cron_task' ) );

class GFPollsSurveysResearchTests extends GFAddOn {

	private static $_instance = null;

	public $_version = GF_POLLS_SURVEYS_RESEARCHTESTS_VERSION;
	protected $_min_gravityforms_version = '1.9.15.12';
	protected $_slug = 'brewpixel-gravityforms-polls-surveys-researchtests';
	protected $_path = 'brewpixel-gravityforms-polls-surveys-researchtests/polls.php';
	protected $_full_path = __FILE__;
	protected $_url = 'https://brewpixel.com';
	protected $_title = 'Gravity Forms Polls, Surveys and Research Tests Add-On';
	protected $_short_title = 'Polls Surveys Research';
	protected $_enable_rg_autoupgrade = true;
	private $_form_meta_by_id = array();
	public $gpollsurveyresearchtest_add_scripts;
	private $_random_ids = array();
	
	/**
	 * Members plugin integration
	 */
	protected $_capabilities = array(
		'brewpixel_gravityforms_polls_surveys_researchtests',
		'brewpixel_gravityforms_polls_surveys_researchtests_uninstall',
		'brewpixel_gravityforms_polls_surveys_researchtests_results',
		'brewpixel_gravityforms_polls_surveys_researchtests_settings',
		'brewpixel_gravityforms_polls_surveys_researchtests_form_settings'
	);

	/**
	 * Permissions
	 */
	protected $_capabilities_settings_page = 'brewpixel_gravityforms_polls_surveys_researchtests_settings';
	protected $_capabilities_form_settings = 'brewpixel_gravityforms_polls_surveys_researchtests_form_settings';
	protected $_capabilities_uninstall = 'brewpixel_gravityforms_polls_surveys_researchtests_uninstall';

	/**
	 * Get an instance of this class.
	 *
	 * @return GFPollsSurveysResearchTests
	 */
	public static function get_instance() {
		if ( self::$_instance == null ) {
			self::$_instance = new GFPollsSurveysResearchTests();
		}

		return self::$_instance;
	}

	private function __clone() {
	} /* do nothing */

	/**
	 * Handles anything which requires early initialization.
	 */
	public function pre_init() {
		parent::pre_init();

		if ( $this->is_gravityforms_supported() && class_exists( 'GF_Field' ) ) {
			require_once( 'includes/class-gf-field-poll-survey-researchtest.php' );

			add_action( 'gform_after_submission', array( $this, 'after_submission' ), 10, 2 );
			add_filter( 'gform_export_field_value', array( $this, 'display_entries_field_value' ), 10, 4 );
			add_action( 'gform_polls_cron', array( $this, 'wp_cron_task' ) );
		}
	}

	/**
	 * Handles hooks and loading of language files.
	 */
	public function init() {

		// add a special class to poll fields so we can identify them later
		add_action( 'gform_field_css_class', array( $this, 'add_custom_class' ), 10, 3 );

		if ( $this->get_custom_cron_schedule() ) {
			add_filter( 'cron_schedules', array( $this, 'cron_add_custom_schedule' ) );
		}

		// add wp_cron job if it's not already scheduled
		if ( ! wp_next_scheduled( 'gform_polls_cron' ) ) {
			wp_schedule_event( time(), $this->get_cron_recurrence(), 'gform_polls_cron' );
		}

		add_filter( 'gform_shortcode_pollssurveysresearchtests', array( $this, 'pollsurveyresearchtest_shortcode' ), 10, 3 );

		add_filter( 'gform_pre_render', array( $this, 'pre_render' ) );

		// shuffle choices if configured
		add_filter( 'gform_field_content', array( $this, 'render_pollsurveyresearchtest_field_content' ), 10, 5 );

		add_action( 'gform_validation', array( $this, 'form_validation' ) );

		// update the cache
		add_action( 'gform_entry_created', array( $this, 'entry_created' ), 10, 2 );

		// maybe display results on confirmation
		add_filter( 'gform_confirmation', array( $this, 'display_confirmation' ), 10, 4 );

		add_shortcode( 'gfpollssurveysresearchtests_total', array( $this, 'pollsurveyresearchtest_total_shortcode' ) );

		// merge tags
		add_filter( 'gform_replace_merge_tags', array( $this, 'render_merge_tag' ), 10, 7 );

		add_filter( 'gform_entry_field_value', array( $this, 'display_poll_on_entry_detail' ), 10, 4 );

		// Integration with the feed add-ons as of GF 1.9.15.12; for add-ons which don't override get_field_value().
		add_filter( 'gform_addon_field_value', array( $this, 'addon_field_value' ), 10, 5 );

		// AWeber 2.3 and newer use the gform_addon_field_value hook, only use the gform_aweber_field_value hook with older versions.
		if ( defined( 'GF_AWEBER_VERSION' ) && version_compare( GF_AWEBER_VERSION, '2.3', '<' ) ) {
			add_filter( 'gform_aweber_field_value', array( $this, 'display_entries_field_value' ), 10, 4 );
		}

		// Mailchimp Add-On integration
		add_filter( 'gform_mailchimp_field_value', array( $this, 'display_entries_field_value' ), 10, 4 );

		// Campaign Monitor Add-On integration
		add_filter( 'gform_campaignmonitor_field_value', array( $this, 'display_entries_field_value' ), 10, 4 );

		// Zapier Add-On integration
		add_filter( 'gform_zapier_field_value', array( $this, 'display_entries_field_value' ), 10, 4 );

		parent::init();
	}

	/**
	 * Initialize the admin specific hooks.
	 */
	public function init_admin() {

		//form editor
		add_action( 'gform_field_standard_settings', array( $this, 'pollsurveyresearchtest_field_settings' ), 10, 2 );
		add_filter( 'gform_tooltips', array( $this, 'add_pollsurveyresearchtest_field_tooltips' ) );
		add_action( 'gform_after_save_form', array( $this, 'after_save_form' ), 10, 2 );

		//display poll results on entry list
		add_filter( 'gform_entries_field_value', array( $this, 'display_entries_field_value' ), 10, 4 );

		//merge tags
		add_filter( 'gform_admin_pre_render', array( $this, 'add_merge_tags' ) );

		//update the cache
		add_action( 'gform_after_update_entry', array( $this, 'entry_updated' ), 10, 2 );
		add_action( 'gform_update_status', array( $this, 'update_entry_status' ), 10, 2 );

		// contacts
		add_filter( 'gform_contacts_tabs_contact_detail', array( $this, 'add_tab_to_contact_detail' ), 10, 2 );
		add_action( 'gform_contacts_tab_polls', array( $this, 'contacts_tab' ) );

		// Adds the polls action to the shortcode builder UI
		add_filter( 'gform_shortcode_builder_actions', array( $this, 'add_pollssurveysresearchtests_shortcode_ui_action' ) );

		parent::init_admin();
	}

	/**
	 * Initialize the AJAX hooks.
	 */
	public function init_ajax() {

		add_action( 'wp_ajax_gpollsurveyresearchtest_ajax', array( $this, 'gpollsurveyresearchtest_ajax' ) );
		add_action( 'wp_ajax_nopriv_gpollsurveyresearchtest_ajax', array( $this, 'gpollsurveyresearchtest_ajax' ) );

		parent::init_ajax();
	}

	/**
	 * The Polls add-on does not support logging.
	 *
	 * @param array $plugins The plugins which support logging.
	 *
	 * @return array
	 */
	public function set_logging_supported( $plugins ) {

		return $plugins;

	}

	// # SCRIPTS & STYLES -----------------------------------------------------------------------------------------------

	/**
	 * Return the scripts which should be enqueued.
	 *
	 * @return array
	 */
	public function scripts() {
		$scripts = array(
			array(
				'handle'  => 'gpollsurveyresearchtest_form_editor_js',
				'src'     => $this->get_base_url() . '/js/gpollsurveyresearchtest_form_editor.js',
				'version' => $this->_version,
				'deps'    => array( 'jquery' ),
				'strings' => array(
					'firstChoice'  => esc_html__( 'First Choice', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
					'secondChoice' => esc_html__( 'Second Choice', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
					'thirdChoice'  => esc_html__( 'Third Choice', 'brewpixel-gravityforms-polls-surveys-researchtests' )
				),
				'enqueue' => array(
					array( 'admin_page' => array( 'form_editor' ) ),
				)
			),
			array(
				'handle'  => 'gpollsurveyresearchtest_form_settings_js',
				'src'     => $this->get_base_url() . '/js/gpollsurveyresearchtest_form_settings.js',
				'version' => $this->_version,
				'deps'    => array( 'jquery' ),
				'enqueue' => array(
					array(
						'admin_page' => array( 'form_settings' ),
						'tab'        => 'brewpixel-gravityforms-polls-surveys-researchtests',
					),
				)
			),
			array(
				'handle'   => 'gpollsurveyresearchtest_js',
				'src'      => $this->get_base_url() . '/js/gpollsurveyresearchtest.js',
				'version'  => $this->_version,
				'deps'     => array( 'jquery' ),
				'callback' => array( $this, 'localize_scripts' ),
				'enqueue'  => array(
					array( 'field_types' => array( 'pollsurveyresearchtest' ) ),
				)
			),
		);

		return array_merge( parent::scripts(), $scripts );
	}

	/**
	 * Return the stylesheets which should be enqueued.
	 *
	 * @return array
	 */
	public function styles() {

		$styles = array(
			array(
				'handle'  => 'gpollsurveyresearchtest_form_editor_css',
				'src'     => $this->get_base_url() . '/css/gpollsurveyresearchtest_form_editor.css',
				'version' => $this->_version,
				'enqueue' => array(
					array( 'admin_page' => array( 'form_editor' ) ),
				)
			),
			array(
				'handle'  => 'gpollsurveyresearchtest_css',
				'src'     => $this->get_base_url() . '/css/gpollsurveyresearchtest.css',
				'version' => $this->_version,
				'enqueue' => array(
					array( 'field_types' => array( 'pollsurveyresearchtest' ) ),
					array( 'admin_page' => array( 'form_editor', 'results', 'entry_view', 'entry_detail' ) ),
				)
			),
		);

		return array_merge( parent::styles(), $styles );
	}

	/**
	 * Localize the strings used by the scripts.
	 */
	public function localize_scripts() {

		// Get current page protocol
		$protocol = isset( $_SERVER['HTTPS'] ) ? 'https://' : 'http://';
		// Output admin-ajax.php URL with same protocol as current page
		$params = array(
			'ajaxurl'   => admin_url( 'admin-ajax.php', $protocol ),
			'imagesUrl' => $this->get_base_url() . '/images'
		);
		wp_localize_script( 'gpollsurveyresearchtest_js', 'gpollsurveyresearchtestVars', $params );

		//localisable strings for the js file
		$strings = array(
			'viewResults'   => esc_html__( 'View results', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
			'backToThePoll' => esc_html__( 'Back to the poll', 'brewpixel-gravityforms-polls-surveys-researchtests' )

		);
		wp_localize_script( 'gpollsurveyresearchtest_js', 'gpollsurveyresearchtest_strings', $strings );

	}


	// # RESULTS --------------------------------------------------------------------------------------------------------

	/**
	 * Configure the survey results page.
	 *
	 * @return array
	 */
	public function get_results_page_config() {
		return array(
			'title'        => esc_html__( 'Poll Results', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
			'capabilities' => array( 'brewpixel_gravityforms_polls_surveys_researchtests_results' ),
			'callbacks'    => array(
				'fields' => array( $this, 'results_fields' )
			)
		);
	}

	/**
	 * Get all the poll fields for the current form.
	 *
	 * @param array $form The current form object.
	 *
	 * @return GF_Field[]
	 */
	public function results_fields( $form ) {
		return GFAPI::get_fields_by_type( $form, array( 'pollsurveyresearchtest' ) );
	}


	// # MERGE TAGS -----------------------------------------------------------------------------------------------------

	/**
	 * Add the result merge tags to the merge tag drop downs in the admin.
	 *
	 * @param array $form The current form.
	 *
	 * @return array
	 */
	public function add_merge_tags( $form ) {
		if ( ! $this->is_form_settings() ) {
			return $form;
		}

		$pollsurveyresearchtest_fields = GFAPI::get_fields_by_type( $form, array( 'pollsurveyresearchtest' ) );

		if ( empty( $pollsurveyresearchtest_fields ) ) {
			return $form;
		}

		foreach ( $pollsurveyresearchtest_fields as $field ) {
			$field_id     = $field->id;
			$field_label  = $field->label;
			$group        = $field->isRequired ? 'required' : 'optional';
			$merge_tags[] = array(
				'group' => $group,
				'label' => $field_label . esc_html__( ': Poll Results', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
				'tag'   => "{gpollsurveyresearchtest:field={$field_id}}"
			);
		}

		$merge_tags[] = array( 'group' => 'other', 'label' => esc_html__( 'All Poll Results', 'brewpixel-gravityforms-polls-surveys-researchtests' ), 'tag' => '{all_poll_results}' );

		?>
		<script type="text/javascript">
			if (window.gform)
				gform.addFilter("gform_merge_tags", "gpollsurveyresearchtest_add_merge_tags");
			function gpollsurveyresearchtest_add_merge_tags(mergeTags, elementId, hideAllFields, excludeFieldTypes, isPrepop, option) {
				if (isPrepop)
					return mergeTags;
				var customMergeTags = <?php echo json_encode( $merge_tags ); ?>;
				jQuery.each(customMergeTags, function (i, customMergeTag) {
					mergeTags[customMergeTag.group].tags.push({tag: customMergeTag.tag, label: customMergeTag.label});
				});

				return mergeTags;
			}
		</script>
		<?php
		//return the form object from the php hook
		return $form;
	}

	/**
	 * Replace the result merge tags.
	 *
	 * @param string $text The current text in which merge tags are being replaced.
	 * @param array $form The current form object.
	 * @param array $entry The current entry object.
	 * @param bool $url_encode Whether or not to encode any URLs found in the replaced value.
	 * @param bool $esc_html Whether or not to encode HTML found in the replaced value.
	 * @param bool $nl2br Whether or not to convert newlines to break tags.
	 * @param string $format The format requested for the location the merge is being used. Possible values: html, text or url.
	 *
	 * @return string
	 */
	public function render_merge_tag( $text, $form, $entry, $url_encode, $esc_html, $nl2br, $format ) {

		if ( empty( $entry ) || empty( $form ) ) {
			return $text;
		}

		$pollsurveyresearchtest_fields = GFAPI::get_fields_by_type( $form, array( 'pollsurveyresearchtest' ) );
		if ( empty ( $pollsurveyresearchtest_fields ) ) {
			return $text;
		}

		$enqueue_scripts = false;
		$form_id         = $form['id'];

		preg_match_all( "/{all_poll_results(:(.*?))?}/", $text, $matches, PREG_SET_ORDER );
		foreach ( $matches as $match ) {

			$full_tag       = $match[0];
			$options_string = isset( $match[2] ) ? $match[2] : '';
			$options        = shortcode_parse_atts( $options_string );

			extract(
				shortcode_atts(
					array(
						'field'       => 0,
						'style'       => 'green',
						'percentages' => 'true',
						'counts'      => 'true',
					), $options
				)
			);
			$percentages     = strtolower( $percentages ) == 'false' ? false : true;
			$counts          = strtolower( $counts ) == 'false' ? false : true;
			$results         = $this->gpollsurveyresearchtest_get_results( $form_id, $field, $style, $percentages, $counts, $entry );
			$results_summary = $results['summary'];
			$new_value       = $results_summary;

			$text = str_replace( $full_tag, $new_value, $text );

			$enqueue_scripts = true;
		}

		preg_match_all( "/\{gpollsurveyresearchtest:(.*?)\}/", $text, $matches, PREG_SET_ORDER );

		foreach ( $matches as $match ) {
			$full_tag = $match[0];

			$options_string = isset( $match[1] ) ? $match[1] : '';
			$options        = shortcode_parse_atts( $options_string );

			extract(
				shortcode_atts(
					array(
						'field'       => 0,
						'style'       => 'green',
						'percentages' => 'true',
						'counts'      => 'true',
					), $options
				)
			);

			$percentages     = strtolower( $percentages ) == 'false' ? false : true;
			$counts          = strtolower( $counts ) == 'false' ? false : true;
			$results         = $this->gpollsurveyresearchtest_get_results( $form_id, $field, $style, $percentages, $counts, $entry );
			$results_summary = $results['summary'];
			$new_value       = $results_summary;

			$text = str_replace( $full_tag, $new_value, $text );

			$enqueue_scripts = true;

		}

		if ( $enqueue_scripts ) {
			wp_enqueue_script( 'jquery' );
			wp_enqueue_style( 'gpollsurveyresearchtest_css' );
			wp_enqueue_style( 'gpollsurveyresearchtest_form_editor_css' );
			wp_enqueue_script( 'gpollsurveyresearchtest_js' );
			$this->localize_scripts();
		}

		return $text;

	}

	// # FORM RENDER & SUBMISSION ---------------------------------------------------------------------------------------

	/**
	 * Helper for retrieving a form setting.
	 *
	 * @param array $form The current form object.
	 * @param string $setting_name The property to be retrieved.
	 *
	 * @return bool|string
	 */
	public function get_form_setting( $form, $setting_name ) {
		if ( false === empty( $form ) ) {
			$settings = $this->get_form_settings( $form );

			// check for legacy form settings from a form exported from a previous version pre-framework
			if ( empty( $settings ) && isset( $form['gpollsurveyresearchtestDisplayResults'] ) ) {
				$this->upgrade_form_settings( $form );
			}

			if ( isset( $settings[ $setting_name ] ) ) {
				$setting_value = $settings[ $setting_name ];
				if ( $setting_value == '1' ) {
					$setting_value = true;
				} elseif ( $setting_value == '0' ) {
					$setting_value = false;
				}

				return $setting_value;
			}
		}

		$setting_value = '';
		//default values
		switch ( $setting_name ) {
			case 'displayResults' :
			case 'showResultsLink' :
			case 'showPercentages' :
			case 'showCounts' :
				$setting_value = true;
				break;
			case 'blockRepeatVoters' :
				$setting_value = false;
				break;
			case 'style' :
				$setting_value = 'green';
				break;
			case 'cookie' :
				$setting_value = '1 month';
				break;
		}

		return $setting_value;
	}

	/**
	 * Updates the form object with the appropriate css classes and if necessary configures the select placeholder.
	 *
	 * @param array $form The for currently being processed for display.
	 *
	 * @return array
	 */
	public function pre_render( $form ) {

		$pollsurveyresearchtest_fields = GFAPI::get_fields_by_type( $form, array( 'pollsurveyresearchtest' ) );
		if ( ! empty ( $pollsurveyresearchtest_fields ) ) {
			$form_css          = 'gpollsurveyresearchtest_enabled';
			$show_results_link = $this->get_form_setting( $form, 'showResultsLink' );

			if ( $show_results_link ) {
				$form_css .= ' gpollsurveyresearchtest_show_results_link';
			}

			$block_repeat_voters = $this->get_form_setting( $form, 'blockRepeatVoters' );

			if ( $block_repeat_voters && rgget( 'gf_page' ) != 'preview' ) {
				$form_css .= ' gpollsurveyresearchtest_block_repeat_voters';
			}

			$form['cssClass'] = empty( $form['cssClass'] ) ? $form_css . ' gpollsurveyresearchtest' : $form_css . ' ' . $form['cssClass'];

			foreach ( $form['fields'] as &$field ) {
				if ( $field->type != 'pollsurveyresearchtest' ) {
					continue;
				}

				if ( $field->get_input_type() == 'select' && empty( $field->placeholder ) ) {
					$field->placeholder = esc_html__( 'Select one', 'brewpixel-gravityforms-polls-surveys-researchtests' );
				}
			}
		}
		
		if ( empty ( $pollsurveyresearchtest_fields ) ) {
			return $form;
		}

		//maybe shuffle fields
		if ( rgars( $form, $this->_slug . '/shuffleFields' ) ) {
			$random_id_groups    = $this->get_random_ids( $form );
			$c             = 0;
			$page_number   = 1;
			$random_fields = array();
			foreach ( $random_id_groups as $random_group => $random_ids ) {

				if( 'select' === $random_group ) {
					foreach ( $random_ids as $random_id ) {
						$random_fields[] = $this->get_field_by_id( $form, $random_id );
					}
				} else if( 'number' === $random_group ) {
					foreach ( $random_ids as $random_id ) {
						$random_fields[] = $this->get_field_by_id( $form, $random_id );
					}
				} else if( 'checkbox' === $random_group ) {
					foreach ( $random_ids as $random_id ) {
						$random_fields[] = $this->get_field_by_id( $form, $random_id );
					}
				} else if( 'radio' === $random_group ) {
					foreach ( $random_ids as $random_id ) {
						$random_fields[] = $this->get_field_by_id( $form, $random_id );
					}
				}
			}
			
			foreach ( $form['fields'] as $key => $field ) {
				if ( $field->type == 'pollsurveyresearchtest' ) {
					$form['fields'][ $key ]             = $random_fields[ $c ++ ];
					$form['fields'][ $key ]->pageNumber = $page_number;
				} elseif ( $field->type == 'page' ) {
					$page_number ++;
				}
			}
		}

		foreach ( $form['fields'] as &$field ) {
			if ( $field->type != 'pollsurveyresearchtest' ) {
				continue;
			}

			if ( $field->inputType == 'select' && ! $field->placeholder ) {
				$field->placeholder = esc_html__( 'Select one', 'brewpixel-gravityforms-polls-surveys-researchtests' );
			}
		}

		return $form;
	}

	/**
	 * Retrieve the specified field.
	 *
	 * @param array $form The current form.
	 * @param int $field_id The ID of the field to be retrieved.
	 *
	 * @return GF_Field
	 */
	public function get_field_by_id( $form, $field_id ) {
		foreach ( $form['fields'] as $field ) {
			if ( $field->id == $field_id ) {
				return $field;
			}
		}
	}

	/**
	 * Return a randomized array containing the field IDs.
	 *
	 * @param array $form The current form.
	 *
	 * @return array
	 */
	public function get_random_ids( $form ) {

		$random_ids = array();
		$random_ids[ 'select' ] = [];
		$random_ids[ 'number' ] = [];
		$random_ids[ 'radio' ] = [];
		$random_ids[ 'checkbox' ] = [];
		
		if ( false === empty( $this->_random_ids ) ) {
			$random_ids = $this->_random_ids;
		} elseif ( rgpost( 'gpollsurveyresearchtest_random_ids' ) ) {
			$random_ids = explode( ',', rgpost( 'gpollsurveyresearchtest_random_ids' ) );
		} else {
			$pollsurveyresearchtest_fields = GFAPI::get_fields_by_type( $form, array( 'pollsurveyresearchtest' ) );
			foreach ( $pollsurveyresearchtest_fields as $pollsurveyresearchtest_field ) {
				if( $pollsurveyresearchtest_field->inputType == 'select' ) {
					$random_ids[ 'select' ][] = $pollsurveyresearchtest_field->id;
				} else if( $pollsurveyresearchtest_field->inputType == 'number' ) {
					$random_ids[ 'number' ][] = $pollsurveyresearchtest_field->id;
                } else if( $pollsurveyresearchtest_field->inputType == 'radio' ) {
					$random_ids[ 'radio' ][] = $pollsurveyresearchtest_field->id;
                } else if( $pollsurveyresearchtest_field->inputType == 'checkbox' ) {
					$random_ids['checkbox'][] = $pollsurveyresearchtest_field->id;
				}
				//$random_ids[] = $pollsurveyresearchtest_field->id;
			}
			shuffle( $random_ids[ 'number' ] );
			shuffle( $random_ids[ 'radio' ] );
			$this->_random_ids = $random_ids;
		}

		/*
		 * return multi-dimensional array
		 * array(2) { [0]=> int(4) [1]=> int(3) }
		 * array(2) { [0]=> int(3) [1]=> int(4) }
		 * 
		 * array { 'select' => array( 3, 4 ), 'radio' => array( 3, 4 ), 'checkbox' => array( 3, 4 ), 'number' => array( 3, 4 ) }
		*/
		//var_dump($random_ids);
		return $random_ids;
	}

	/**
	 * If the fields were shuffled append a hidden field after the form tag containing the IDs.
	 *
	 * @param string $form_tag The form tag for the current form.
	 * @param array $form The current form.
	 *
	 * @return string
	 */
	public function maybe_store_selected_field_ids( $form_tag, $form ) {
		if ( $this->get_form_setting( $form, 'shuffleFields' ) ) {
			$value = implode( ',', $this->get_random_ids( $form ) );
			$input = "<input type='hidden' value='$value' name='gpollsurveyresearchtest_random_ids'>";
			$form_tag .= $input;
		}

		return $form_tag;
	}


	/**
	 * If necessary perform validation to prevent repeat voting.
	 *
	 * @param array $validation_result Contains the validation result, the form object, and the failed validation page number.
	 *
	 * @return array
	 */
	public function form_validation( $validation_result ) {
		$form        = $validation_result['form'];
		$pollsurveyresearchtest_fields = GFAPI::get_fields_by_type( $form, array( 'pollsurveyresearchtest' ) );
		if ( empty ( $pollsurveyresearchtest_fields ) ) {
			return $validation_result;
		}
		$form_setting_block_repeat = $this->get_form_setting( $form, 'blockRepeatVoters' );
		$field_values              = wp_parse_args( rgpost( 'gform_field_values' ) );
		if ( $form_setting_block_repeat || ( isset( $field_values['gpollsurveyresearchtest_enabled'] ) && $field_values['gpollsurveyresearchtest_enabled'] == '1' && isset( $field_values['gpollsurveyresearchtest_cookie'] ) && false === empty( $field_values['gpollsurveyresearchtest_cookie'] ) ) ) {
			$form_id = rgar( $form, 'id' );
			if ( isset ( $_COOKIE[ 'gpollsurveyresearchtest_form_' . $form_id ] ) ) {
				// set the form validation to false
				$validation_result['is_valid'] = false;
				foreach ( $form['fields'] as &$field ) {
					if ( $field->type == 'pollsurveyresearchtest' ) {
						$field->failed_validation  = true;
						$field->validation_message = esc_html__( 'Repeat voting is not allowed', 'brewpixel-gravityforms-polls-surveys-researchtests' );
					}
				}
				$validation_result['form'] = $form;
			}
		}

		return $validation_result;
	}

	/**
	 * If necessary randomize the field choices.
	 *
	 * @param string $content The field content to be filtered.
	 * @param GF_Field $field The field currently being processed for display.
	 * @param string|array $value The field value. From default/dynamic population, $_POST, or a resumed incomplete submission.
	 * @param int $entry_id The ID of the entry if the field is being displayed on the entry detail page.
	 * @param int $form_id The ID of the current form.
	 *
	 * @return string
	 */
	public function render_pollsurveyresearchtest_field_content( $content, $field, $value, $entry_id, $form_id ) {

		if ( $entry_id === 0 && $field->type == 'pollsurveyresearchtest' && ! $this->is_form_editor() ) {

			if ( $field->enableRandomizeChoices ) {

				//pass the HTML for the choices through DOMdocument to make sure we get the complete node
				$dom     = new DOMDocument();
				$content = '<?xml version="1.0" encoding="UTF-8"?>' . $content;
				$loader  = libxml_disable_entity_loader( true );
				$errors  = libxml_use_internal_errors( true );
				$dom->loadHTML( $content );
				libxml_clear_errors();
				libxml_use_internal_errors( $errors );
				libxml_disable_entity_loader( $loader );
				$content = $dom->saveXML( $dom->documentElement );

				//pick out the elements: LI for radio & checkbox, OPTION for select
				$element_name = $field->inputType == 'select' ? 'select' : 'ul';
				$nodes        = $dom->getElementsByTagName( $element_name )->item( 0 )->childNodes;

				//cycle through the LI elements and swap them around randomly
				$temp_str1 = 'gpollsurveyresearchtest_shuffle_placeholder1';
				$temp_str2 = 'gpollsurveyresearchtest_shuffle_placeholder2';
				for ( $i = $nodes->length - 1; $i >= 0; $i -- ) {
					$n = rand( 0, $i );
					if ( $i <> $n ) {
						$i_str   = $dom->saveXML( $nodes->item( $i ) );
						$n_str   = $dom->saveXML( $nodes->item( $n ) );
						$content = str_replace( $i_str, $temp_str1, $content );
						$content = str_replace( $n_str, $temp_str2, $content );
						$content = str_replace( $temp_str2, $i_str, $content );
						$content = str_replace( $temp_str1, $n_str, $content );
					}
				}
				//snip off the tags that DOMdocument adds
				$content = str_replace( '<html><body>', '', $content );
				$content = str_replace( '</body></html>', '', $content );
			}
		}

		return $content;
	}

	/**
	 * If necessary update the confirmation to include the results.
	 *
	 * @param string|array $confirmation The forms current confirmation.
	 * @param array $form The form currently being processed.
	 * @param array $lead The entry currently being processed.
	 * @param bool $ajax Indicates if AJAX is enabled for this form.
	 *
	 * @return string|array
	 */
	public function display_confirmation( $confirmation, $form, $lead, $ajax ) {
		$pollsurveyresearchtest_fields = GFAPI::get_fields_by_type( $form, array( 'pollsurveyresearchtest' ) );

		if ( empty ( $pollsurveyresearchtest_fields ) ) {
			return $confirmation;
		}

		$form_id              = $form['id'];
		$display_confirmation = false;
		$display_results      = false;
		$override             = false;

		$field_values = array();
		if ( isset( $_POST['gform_field_values'] ) ) {
			$field_values = wp_parse_args( $_POST['gform_field_values'] );
		}

		// shortcode attributes override form settings
		if ( rgar( $field_values, 'gpollsurveyresearchtest_enabled' ) == '1' ) {

			$field_values      = wp_parse_args( $_POST['gform_field_values'] );
			$show_results_link = rgar( $field_values, 'gpollsurveyresearchtest_show_results_link' );
			$show_results_link = $show_results_link == '1' ? true : false;
			$style             = rgar( $field_values, 'gpollsurveyresearchtest_style' );
			$percentages       = rgar( $field_values, 'gpollsurveyresearchtest_percentages' );
			$percentages       = $percentages == '1' ? true : false;
			$counts            = rgar( $field_values, 'gpollsurveyresearchtest_counts' );
			$counts            = $counts == '1' ? true : false;
			$cookie            = rgar( $field_values, 'gpollsurveyresearchtest_cookie' );

			$display_results = rgar( $field_values, 'gpollsurveyresearchtest_display_results' );
			$display_results = $display_results == '1' ? true : false;

			$display_confirmation = rgar( $field_values, 'gpollsurveyresearchtest_confirmation' );
			$display_confirmation = $display_confirmation == '1' ? true : false;

			$checksum = rgar( $field_values, 'gpollsurveyresearchtest_checksum' );
			if ( $checksum == $this->generate_checksum( $display_results, $show_results_link, $cookie, $display_confirmation, $percentages, $counts, $style ) ) {
				$override = true;
			}

		}


		if ( false === $override ) {
			$style       = $this->get_form_setting( $form, 'style' );
			$percentages = $this->get_form_setting( $form, 'showPercentages' );
			$counts      = $this->get_form_setting( $form, 'showCounts' );

			$display_results      = $this->get_form_setting( $form, 'displayResults' );
			$display_confirmation = true;
		}

		$submitted_fields = array();
		foreach ( $pollsurveyresearchtest_fields as $field ) {
			$field_id    = $field->id;
			$entry_value = RGFormsModel::get_lead_field_value( $lead, $field );
			if ( is_array( $entry_value ) ) {
				$entry_value = implode( '', $entry_value );
			}
			if ( false === empty( $entry_value ) ) {
				$submitted_fields[] = $field_id;
			}
		}

		if ( $display_confirmation && $display_results ) {
			//confirmation message plus results

			//override in the case of headers already sent or ajax = true
			if ( is_array( $confirmation ) && array_key_exists( 'redirect', $confirmation ) ) {
				$confirmation = '';
			}

			//override confirmation if it's a redirect
			$str_pos = strpos( $confirmation, 'gformRedirect' );
			if ( false !== $str_pos ) {
				$confirmation = '';
			}

			$has_confirmation_wrapper = false !== strpos( $confirmation, 'gform_confirmation_wrapper' ) ? true : false;

			if ( $has_confirmation_wrapper ) {
				$confirmation = substr( $confirmation, 0, strlen( $confirmation ) - 6 );
			} //remove the closing div of the wrapper

			$has_confirmation_message = false !== strpos( $confirmation, 'gform_confirmation_message' ) ? true : false;

			if ( $has_confirmation_message ) {
				$confirmation = substr( $confirmation, 0, strlen( $confirmation ) - 6 );
			} //remove the closing div of the message
			else {
				$confirmation .= "<div id='gforms_confirmation_message' class='gform_confirmation_message_{$form_id}'>";
			}

			//$results = $this->gpollsurveyresearchtest_get_results( $form['id'], $submitted_fields, $style, $percentages, $counts, $lead );
			$results = $this->gpollsurveyresearchtest_get_results( $form['id'], 0, $style, $percentages, $counts, $lead );

			$confirmation .= $results['summary'] . '</div>';

			if ( $has_confirmation_wrapper ) {
				$confirmation .= '</div>';
			}

		} elseif ( ! $display_confirmation && $display_results ) {

			//only the results without the confirmation message
			//$results = $this->gpollsurveyresearchtest_get_results( $form['id'], $submitted_fields, $style, $percentages, $counts, $lead );
			$results = $this->gpollsurveyresearchtest_get_results( $form['id'], 0, $style, $percentages, $counts, $lead );

			$results_summary = $results['summary'];
			$confirmation    = sprintf( "<div id='gforms_confirmation_message' class='gform_confirmation_message_{$form_id}'>%s</div>", $results_summary );

		} elseif ( ! $display_confirmation && ! $display_results ) {
			$confirmation = "<div id='gforms_confirmation_message' class='gform_confirmation_message_{$form_id}'></div>";
		}

		return $confirmation;
	} // end function gpollsurveyresearchtest_confirmation

	/**
	 * If necessary set a cookie to block repeat submissions.
	 *
	 * @param array $entry The current entry object.
	 * @param array $form The current form object.
	 */
	public function after_submission( $entry, $form ) {
		if ( rgget( 'gf_page' ) == 'preview' ) {
			return;
		}

		$set_cookie   = false;
		$cookie       = '';
		$field_values = array();
		if ( isset( $_POST['gform_field_values'] ) ) {
			$field_values = wp_parse_args( $_POST['gform_field_values'] );
		}

		$override = false;

		if ( rgar( $field_values, 'gpollsurveyresearchtest_enabled' ) == '1' ) {
			$show_results_link = rgar( $field_values, 'gpollsurveyresearchtest_show_results_link' );
			$show_results_link = $show_results_link == '1' ? true : false;
			$style             = rgar( $field_values, 'gpollsurveyresearchtest_style' );
			$percentages       = rgar( $field_values, 'gpollsurveyresearchtest_percentages' );
			$percentages       = $percentages == '1' ? true : false;
			$counts            = rgar( $field_values, 'gpollsurveyresearchtest_counts' );
			$counts            = $counts == '1' ? true : false;
			$cookie            = $field_values['gpollsurveyresearchtest_cookie'];

			$display_results = rgar( $field_values, 'gpollsurveyresearchtest_display_results' );
			$display_results = $display_results == '1' ? true : false;

			$display_confirmation = rgar( $field_values, 'gpollsurveyresearchtest_confirmation' );
			$display_confirmation = $display_confirmation == '1' ? true : false;

			$checksum = rgar( $field_values, 'gpollsurveyresearchtest_checksum' );
			if ( $checksum == $this->generate_checksum( $display_results, $show_results_link, $cookie, $display_confirmation, $percentages, $counts, $style ) ) {
				$set_cookie = true;
				$override   = true;
			}
		}

		if ( false === $override ) {
			if ( $this->get_form_setting( $form, 'blockRepeatVoters' ) ) {
				$set_cookie = true;
				$cookie     = $this->get_form_setting( $form, 'cookie' );
			}
		}

		if ( $set_cookie ) {
			$form_id    = $form['id'];
			$lead_id    = $entry['id'];
			$server_tz  = date_default_timezone_get();
			$browser_tz = rgar( $_COOKIE, 'gpollsurveyresearchtest-timezone' ); // in hours
			if ( false === empty( $browser_tz ) ) {
				date_default_timezone_set( $browser_tz );
			}
			$cookie_expiration = strtotime( $cookie );
			date_default_timezone_set( $server_tz );
			setcookie( 'gpollsurveyresearchtest_form_' . $form_id, $lead_id, $cookie_expiration, COOKIEPATH, COOKIE_DOMAIN );
		}
	}


	// # ENTRY RELATED --------------------------------------------------------------------------------------------------

	/**
	 * If the field is a Poll type radio, select or checkbox then replace the choice value with the choice text.
	 *
	 * @param string $value The field value.
	 * @param GF_Field|null $field The field object being processed or null.
	 *
	 * @return string
	 */
	public function maybe_format_field_values( $value, $field ) {

		if ( is_object( $field ) && $field->type == 'pollsurveyresearchtest' ) {
			switch ( $field->inputType ) {
				case 'radio' :
				case 'select' :
					return RGFormsModel::get_choice_text( $field, $value );

				case 'checkbox' :
					if ( is_array( $value ) ) {
						foreach ( $value as &$choice ) {
							if ( ! empty( $choice ) ) {
								$choice = RGFormsModel::get_choice_text( $field, $choice );
							}
						}
					} else {
						foreach ( $field->choices as $choice ) {
							$val   = rgar( $choice, 'value' );
							$text  = rgar( $choice, 'text' );
							$value = str_replace( $val, $text, $value );
						}
					}
			}
		}

		return $value;
	}

	/**
	 * Format the Poll field values so they use the choice text instead of values before being passed to the third-party.
	 *
	 * @param string $value The field value.
	 * @param array $form The form currently being processed.
	 * @param array $entry The entry object currently being processed.
	 * @param string $field_id The ID of the field currently being processed.
	 *
	 * @return string
	 */
	public function addon_field_value( $value, $form, $entry, $field_id, $slug ) {
		if ( ! empty( $value ) ) {
			$field = RGFormsModel::get_field( $form, $field_id );

			return $this->maybe_format_field_values( $value, $field );
		}

		return $value;
	}

	/**
	 * Format the Poll field values so they use the choice text instead of values.
	 *
	 * Used for the entry list page, the AWeber, Campaign Monitor, and MailChimp add-ons.
	 *
	 * @param string|array $value The field value.
	 * @param int $form_id The ID of the form currently being processed.
	 * @param string $field_id The ID of the field currently being processed.
	 * @param array $entry The entry object currently being processed.
	 *
	 * @return string|array
	 */
	public function display_entries_field_value( $value, $form_id, $field_id, $entry ) {
		if ( ! empty( $value ) ) {
			global $_form_metas;

			if ( ! isset( $_form_metas[ $form_id ] ) ) {
				$_form_metas[ $form_id ] = RGFormsModel::get_form_meta( $form_id );
			}

			$form_meta = $_form_metas[ $form_id ];

			$field = RGFormsModel::get_field( $form_meta, $field_id );

			return $this->maybe_format_field_values( $value, $field );
		}

		return $value;
	}

	/**
	 * Format the Poll field values for display on the entry detail page and print entry.
	 *
	 * @param string|array $value The field value.
	 * @param GF_Field $field The field currently being processed.
	 * @param array $entry The entry object currently being processed.
	 * @param array $form The form object currently being processed.
	 *
	 * @return string|array
	 */
	public function display_poll_on_entry_detail( $value, $field, $entry, $form ) {

		if ( $field->type == 'pollsurveyresearchtest' ) {
			if ( $field->is_entry_detail() ) {
				$results                 = $this->gpollsurveyresearchtest_get_results( $form['id'], $field->id, 'green', true, true, $entry );
				$new_value               = sprintf( '<div class="gpollsurveyresearchtest_entry">%s</div>', rgar( $results, 'summary' ) );
				$this->gpollsurveyresearchtest_add_scripts = true;

				//if original response is not in results display below
				$selected_values  = $this->get_selected_values( $form['id'], $field->id, $entry );
				$possible_choices = $this->get_possible_choices( $form['id'], $field->id );
				foreach ( $selected_values as $selected_value ) {
					if ( ! in_array( $selected_value, $possible_choices ) ) {
						$new_value = sprintf( '%s<h2>%s</h2>%s', $new_value, esc_html__( 'Original Response', 'brewpixel-gravityforms-polls-surveys-researchtests' ), $value );
						break;
					}
				}

				return $new_value;
			} elseif ( is_array( $field->choices ) ) {
				if ( $field->inputType == 'checkbox' ) {
					foreach ( $field->choices as $choice ) {
						$val   = rgar( $choice, 'value' );
						$text  = rgar( $choice, 'text' );
						$value = str_replace( $val, $text, $value );
					}
				} else {
					$value = RGFormsModel::get_choice_text( $field, $value );
				}
			}
		}

		return $value;
	}

	/**
	 * Maybe update the poll results cache when the entry status is changed.
	 *
	 * @param int $entry_id The ID of the entry which was updated.
	 */
	public function update_entry_status( $entry_id ) {
		$entry       = RGFormsModel::get_lead( $entry_id );
		$form_id     = $entry['form_id'];
		$form        = GFFormsModel::get_form_meta( $form_id );
		$pollsurveyresearchtest_fields = GFAPI::get_fields_by_type( $form, array( 'pollsurveyresearchtest' ) );
		if ( empty ( $pollsurveyresearchtest_fields ) ) {
			return;
		}

		$this->maybe_update_cache( $form_id );
	}

	/**
	 * Maybe update the poll results cache when the entry is updated.
	 *
	 * @param array $form The current form object.
	 * @param int $entry_id The ID of the entry which was updated.
	 */
	public function entry_updated( $form, $entry_id ) {
		$pollsurveyresearchtest_fields = GFAPI::get_fields_by_type( $form, array( 'pollsurveyresearchtest' ) );
		if ( empty ( $pollsurveyresearchtest_fields ) ) {
			return;
		}

		$form_id = $form['id'];
		$this->maybe_update_cache( $form_id );
	}

	// # TO FRAMEWORK MIGRATION -----------------------------------------------------------------------------------------

	/**
	 * Checks if a previous version was installed and if the form settings need migrating to the framework structure.
	 *
	 * @param string $previous_version The version number of the previously installed version.
	 */
	public function upgrade( $previous_version ) {
		$previous_is_pre_addon_framework = version_compare( $previous_version, '1.5.4', '<' );

		if ( $previous_is_pre_addon_framework ) {
			$forms = GFFormsModel::get_forms();
			foreach ( $forms as $form ) {
				$form_meta = GFFormsModel::get_form_meta( $form->id );
				$this->upgrade_form_settings( $form_meta );
			}
		}
	}

	/**
	 * Migrates the polls related form settings to the new structure.
	 *
	 * @param array $form The form object currently being processed.
	 */
	private function upgrade_form_settings( $form ) {
		if ( false === isset( $form['gpollsurveyresearchtestDisplayResults'] ) ) {
			return;
		}

		$legacy_form_settings = array(
			'gpollsurveyresearchtestDisplayResults'    => 'displayResults',
			'gpollsurveyresearchtestShowResultsLink'   => 'showResultsLink',
			'gpollsurveyresearchtestShowPercentages'   => 'showPercentages',
			'gpollsurveyresearchtestShowCounts'        => 'showCounts',
			'gpollsurveyresearchtestBlockRepeatVoters' => 'blockRepeatVoters',
			'gpollsurveyresearchtestStyle'             => 'style',
			'gpollsurveyresearchtestCookie'            => 'cookie',
		);

		$new_settings = array();
		foreach ( $legacy_form_settings as $legacy_key => $new_key ) {
			if ( isset( $form[ $legacy_key ] ) ) {
				$new_settings[ $new_key ] = $form[ $legacy_key ];
				unset( $form[ $legacy_key ] );
			}
		}
		if ( false === empty( $new_settings ) ) {
			$form[ $this->_slug ] = $new_settings;
			GFFormsModel::update_form_meta( $form['id'], $form );
		}
	}


	// # FORM SETTINGS --------------------------------------------------------------------------------------------------

	/**
	 * Add the form settings tab.
	 *
	 * @param array $tabs The tabs to be displayed on the form settings page.
	 * @param int $form_id The ID of the current form.
	 *
	 * @return array
	 */
	public function add_form_settings_menu( $tabs, $form_id ) {
		$form        = $this->get_form_meta( $form_id );
		$pollsurveyresearchtest_fields = GFAPI::get_fields_by_type( $form, array( 'pollsurveyresearchtest' ) );
		if ( ! empty( $pollsurveyresearchtest_fields ) ) {
			$tabs[] = array( 'name' => 'brewpixel-gravityforms-polls-surveys-researchtests', 'label' => esc_html__( 'Polls', 'brewpixel-gravityforms-polls-surveys-researchtests' ) );
		}

		return $tabs;
	}

	/**
	 * The settings fields to be rendered on the form settings page.
	 *
	 * @param array $form The current form object.
	 *
	 * @return array
	 */
	public function form_settings_fields( $form ) {

		// check for legacy form settings from a form exported from a previous version pre-framework
		$page = rgget( 'page' );
		if ( 'gf_edit_forms' == $page && false === empty( $form_id ) ) {
			$settings = $this->get_form_settings( $form );
			if ( empty( $settings ) && isset( $form['gpollsurveyresearchtestDisplayResults'] ) ) {
				$this->upgrade_form_settings( $form );
			}
		}

		return array(
			array(
				'title'  => esc_html__( 'Poll Settings', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
				'fields' => array(
					array(
						'name'    => 'general',
						'label'   => esc_html__( 'General', 'gravityformquiz' ),
						'type'    => 'checkbox',
						'choices' => array(
							0 => array(
								'label'         => esc_html__( 'Shuffle test research/poll/survey fields', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
								'name'          => 'shuffleFields',
								'default_value' => $this->get_form_setting( array(), 'shuffleFields' ),
								'tooltip'       => '<h6>' . esc_html__( 'Shuffle Fields', 'brewpixel-gravityforms-polls-surveys-researchtests' ) . '</h6>' . esc_html__( "Display the quiz fields in a random order. This doesn't affect the position of the other fields on the form.", 'brewpixel-gravityforms-polls-surveys-researchtests' )
							),
						)
					),
					array(
						'label'   => esc_html__( 'Results', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
						'type'    => 'checkbox',
						'name'    => 'displayResults',
						'tooltip' => '<h6>' . esc_html__( 'Results', 'brewpixel-gravityforms-polls-surveys-researchtests' ) . '</h6>' . esc_html__( 'Select this option to display the results of submitted poll fields after the form is submitted.', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
						'choices' => array(
							0 => array(
								'label'         => esc_html__( 'Display results of submitted poll fields after voting', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
								'name'          => 'displayResults',
								'default_value' => $this->get_form_setting( array(), 'displayResults' )
							)
						)
					),
					array(
						'label'   => esc_html__( 'Results Link', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
						'type'    => 'checkbox',
						'name'    => 'showResultsLink',
						'tooltip' => '<h6>' . esc_html__( 'Results Link', 'brewpixel-gravityforms-polls-surveys-researchtests' ) . '</h6>' . esc_html__( 'Select this option to add a link to the form which allows the visitor to see the results without voting.', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
						'choices' => array(
							0 => array(
								'label'         => esc_html__( 'Add a poll results link to the form', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
								'name'          => 'showResultsLink',
								'default_value' => $this->get_form_setting( array(), 'showResultsLink' ),
							)
						)
					),
					array(
						'label'   => esc_html__( 'Percentages', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
						'type'    => 'checkbox',
						'name'    => 'showPercentages',
						'tooltip' => '<h6>' . esc_html__( 'Show Percentages', 'brewpixel-gravityforms-polls-surveys-researchtests' ) . '</h6>' . esc_html__( 'Show the percentage of the total votes for each choice.', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
						'choices' => array(
							0 => array(
								'label'         => esc_html__( 'Display percentages', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
								'name'          => 'showPercentages',
								'default_value' => $this->get_form_setting( array(), 'showPercentages' )
							)
						)
					),
					array(
						'label'   => esc_html__( 'Counts', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
						'type'    => 'checkbox',
						'name'    => 'showCounts',
						'tooltip' => '<h6>' . esc_html__( 'Show Counts', 'brewpixel-gravityforms-polls-surveys-researchtests' ) . '</h6>' . esc_html__( 'Show the total number of votes for each choice.', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
						'choices' => array(
							0 => array(
								'label'         => esc_html__( 'Display counts', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
								'name'          => 'showCounts',
								'default_value' => $this->get_form_setting( array(), 'showCounts' )
							)
						)
					),
					array(
						'label'         => esc_html__( 'Style', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
						'name'          => 'style',
						'default_value' => $this->get_form_setting( array(), 'style' ),
						'type'          => 'select',
						'choices'       => array(
							array( 'label' => esc_html__( 'Green', 'brewpixel-gravityforms-polls-surveys-researchtests' ), 'value' => 'green' ),
							array( 'label' => esc_html__( 'Blue', 'brewpixel-gravityforms-polls-surveys-researchtests' ), 'value' => 'blue' ),
							array( 'label' => esc_html__( 'Red', 'brewpixel-gravityforms-polls-surveys-researchtests' ), 'value' => 'red' ),
							array( 'label' => esc_html__( 'Orange', 'brewpixel-gravityforms-polls-surveys-researchtests' ), 'value' => 'orange' ),
						)
					),
					array(
						'label'   => esc_html__( 'Block repeat voters', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
						'type'    => 'block_repeat_voters',
						'name'    => 'blockRepeatVoters',
						'tooltip' => '<h6>' . esc_html__( 'Block Repeat Voters', 'brewpixel-gravityforms-polls-surveys-researchtests' ) . '</h6>' . esc_html__( "Choose whether to allow visitors to vote more than once. Repeat voting is controlled by storing a cookie on the visitor's computer.", 'brewpixel-gravityforms-polls-surveys-researchtests' ),
					),
				)
			),
		);
	}

	/**
	 * Define the properties and output the markup for the block_repeat_voters field type.
	 */
	public function settings_block_repeat_voters() {
		$this->settings_radio(
			array(
				'label'         => esc_html__( 'Repeat Voters', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
				'name'          => 'blockRepeatVoters',
				'class'         => 'gpollsurveyresearchtest-block-repeat-voters',
				'default_value' => $this->get_form_setting( array(), 'blockRepeatVoters' ) ? '1' : '0',
				'type'          => 'radio',
				'choices'       => array(
					array( 'label' => esc_html__( "Don't block repeat voting", 'brewpixel-gravityforms-polls-surveys-researchtests' ), 'value' => '0' ),
					array( 'label' => esc_html__( 'Block repeat voting using cookie', 'brewpixel-gravityforms-polls-surveys-researchtests' ), 'value' => '1' ),
				)
			)
		);

		esc_html_e( 'Expires: ', 'brewpixel-gravityforms-polls-surveys-researchtests' );
		$this->settings_select(
			array(
				'name'          => 'cookie',
				'default_value' => $this->get_form_setting( array(), 'cookie' ),
				'type'          => 'select',
				'choices'       => array(
					array( 'label' => esc_html__( 'Never', 'brewpixel-gravityforms-polls-surveys-researchtests' ), 'value' => '20 years' ),
					array( 'label' => esc_html__( '1 hour', 'brewpixel-gravityforms-polls-surveys-researchtests' ), 'value' => '1 hour' ),
					array( 'label' => esc_html__( '6 hours', 'brewpixel-gravityforms-polls-surveys-researchtests' ), 'value' => '6 hours' ),
					array( 'label' => esc_html__( '12 hours', 'brewpixel-gravityforms-polls-surveys-researchtests' ), 'value' => '12 hours' ),
					array( 'label' => esc_html__( '1 day', 'brewpixel-gravityforms-polls-surveys-researchtests' ), 'value' => '1 day' ),
					array( 'label' => esc_html__( '1 week', 'brewpixel-gravityforms-polls-surveys-researchtests' ), 'value' => '1 week' ),
					array( 'label' => esc_html__( '1 month', 'brewpixel-gravityforms-polls-surveys-researchtests' ), 'value' => '1 month' ),
				)
			)
		);
	}


	// # FIELD SETTINGS -------------------------------------------------------------------------------------------------

	/**
	 * Add the gpollsurveyresearchtest_field class to the Poll field.
	 *
	 * @param string $classes The CSS classes to be filtered, separated by empty spaces.
	 * @param GF_Field $field The field currently being processed.
	 * @param array $form The form currently being processed.
	 *
	 * @return string
	 */
	public function add_custom_class( $classes, $field, $form ) {
		if ( $field->type == 'pollsurveyresearchtest' ) {
			$classes .= ' gpollsurveyresearchtest_field';
		}

		return $classes;
	}

	/**
	 * Add the tooltips for the Poll field.
	 *
	 * @param array $tooltips An associative array of tooltips where the key is the tooltip name and the value is the tooltip.
	 *
	 * @return array
	 */
	public function add_pollsurveyresearchtest_field_tooltips( $tooltips ) {
		//form settings
		$tooltips['gpollsurveyresearchtest_form_settings_display_results']      = '<h6>' . esc_html__( 'Display Results After Voting', 'brewpixel-gravityforms-polls-surveys-researchtests' ) . '</h6>' . esc_html__( 'Select this to display the results of submitted poll fields.', 'brewpixel-gravityforms-polls-surveys-researchtests' );
		$tooltips['gpollsurveyresearchtest_form_settings_display_confirmation'] = '<h6>' . esc_html__( 'Display Confirmation', 'brewpixel-gravityforms-polls-surveys-researchtests' ) . '</h6>' . esc_html__( 'Select this option to display the form confirmation message after the visitor has voted.', 'brewpixel-gravityforms-polls-surveys-researchtests' );
		$tooltips['gpollsurveyresearchtest_form_settings_show_results_link']    = '<h6>' . esc_html__( 'Show Results Link', 'brewpixel-gravityforms-polls-surveys-researchtests' ) . '</h6>' . esc_html__( 'Add a link to the form which allows the visitor to see the results without voting.', 'brewpixel-gravityforms-polls-surveys-researchtests' );
		$tooltips['gpollsurveyresearchtest_form_settings_show_percentages']     = '<h6>' . esc_html__( 'Show Percentages', 'brewpixel-gravityforms-polls-surveys-researchtests' ) . '</h6>' . esc_html__( 'Show the percentage of the total votes for each choice.', 'brewpixel-gravityforms-polls-surveys-researchtests' );
		$tooltips['gpollsurveyresearchtest_form_settings_show_counts']          = '<h6>' . esc_html__( 'Show Counts', 'brewpixel-gravityforms-polls-surveys-researchtests' ) . '</h6>' . esc_html__( 'Show the total number of votes for each choice.', 'brewpixel-gravityforms-polls-surveys-researchtests' );
		$tooltips['gpollsurveyresearchtest_form_settings_repeat_voters']        = '<h6>' . esc_html__( 'Repeat Voters', 'brewpixel-gravityforms-polls-surveys-researchtests' ) . '</h6>' . esc_html__( "Choose whether to allow visitors to vote more than once. Repeat voting is controlled by storing a cookie on the visitor's computer.", 'brewpixel-gravityforms-polls-surveys-researchtests' );

		$tooltips['form_pollsurveyresearchtest_question']           = '<h6>' . esc_html__( 'Poll Question', 'brewpixel-gravityforms-polls-surveys-researchtests' ) . '</h6>' . esc_html__( 'Enter the question you would like to ask the user. The user can then answer the question by selecting from the available choices.', 'brewpixel-gravityforms-polls-surveys-researchtests' );
		$tooltips['form_pollsurveyresearchtest_field_type']         = '<h6>' . esc_html__( 'Poll Type', 'brewpixel-gravityforms-polls-surveys-researchtests' ) . '</h6>' . esc_html__( "Select the field type you'd like to use for the poll.", 'brewpixel-gravityforms-polls-surveys-researchtests' );
		$tooltips['form_field_randomize_choices'] = '<h6>' . esc_html__( 'Randomize Choices', 'brewpixel-gravityforms-polls-surveys-researchtests' ) . '</h6>' . esc_html__( 'Check the box to randomize the order in which the choices are displayed to the user. This setting affects only voting - it will not affect the order of the results.', 'brewpixel-gravityforms-polls-surveys-researchtests' );

		return $tooltips;
	}

	/**
	 * Add the custom settings for the Poll fields to the fields general tab.
	 *
	 * @param int $position The position the settings should be located at.
	 * @param int $form_id The ID of the form currently being edited.
	 */
	public function pollsurveyresearchtest_field_settings( $position, $form_id ) {

		//create settings on position 25 (right after Field Label)
		if ( $position == 25 ) {
			?>

			<li class="pollsurveyresearchtest_question_setting field_setting">
				<label for="pollsurveyresearchtest_question" class="section_label">
					<?php esc_html_e( 'Poll Question', 'brewpixel-gravityforms-polls-surveys-researchtests' ); ?>
					<?php gform_tooltip( 'form_pollsurveyresearchtest_question' ); ?>
				</label>
				<input type="text" id="pollsurveyresearchtest_question" class="fieldwidth-3" onkeyup="SetFieldLabel(this.value)"
				       size="35"/>
			</li>

			<li class="pollsurveyresearchtest_field_type_setting field_setting">
				<label for="pollsurveyresearchtest_field_type" class="section_label">
					<?php esc_html_e( 'Poll Type', 'brewpixel-gravityforms-polls-surveys-researchtests' ); ?>
					<?php gform_tooltip( 'form_pollsurveyresearchtest_field_type' ); ?>
				</label>
				<select id="pollsurveyresearchtest_field_type"
				        onchange="if(jQuery(this).val() == '') return; jQuery('#field_settings').slideUp(function(){StartChangePollType(jQuery('#pollsurveyresearchtest_field_type').val());});">
					<option value="select"><?php esc_html_e( 'Drop Down', 'brewpixel-gravityforms-polls-surveys-researchtests' ); ?></option>
					<option value="radio"><?php esc_html_e( 'Radio Buttons', 'brewpixel-gravityforms-polls-surveys-researchtests' ); ?></option>
					<option value="checkbox"><?php esc_html_e( 'Checkboxes', 'brewpixel-gravityforms-polls-surveys-researchtests' ); ?></option>
					<option value="number"><?php esc_html_e( 'Number', 'brewpixel-gravityforms-polls-surveys-researchtests' ); ?></option>
					
				</select>

			</li>

			<?php
		} elseif ( $position == 1368 ) {
			//right after the other_choice_setting
			?>
			<li class="randomize_choices_setting field_setting">

				<input type="checkbox" id="field_randomize_choices"
				       onclick="var value = jQuery(this).is(':checked'); SetFieldProperty('enableRandomizeChoices', value); UpdateFieldChoices(GetInputType(field));"/>
				<label for="field_randomize_choices" class="inline">
					<?php esc_html_e( 'Randomize order of choices', 'brewpixel-gravityforms-polls-surveys-researchtests' ); ?>
					<?php gform_tooltip( 'form_field_randomize_choices' ) ?>
				</label>

			</li>
			<?php
		}
	}


	// # AJAX FUNCTIONS ------------------------------------------------------------------------------------------------

	/**
	 * Handler for the gpollsurveyresearchtest_ajax AJAX request.
	 * Returns the json encoded result for processing by gpollsurveyresearchtest.js.
	 */
	public function gpollsurveyresearchtest_ajax() {
		$output = array();

		$form_id = rgpost( 'formId' );
		$form    = RGFormsModel::get_form_meta( $form_id );

		$preview_results = rgpost( 'previewResults' );
		$preview_results = $preview_results == '1' ? true : false;

		$has_voted = isset ( $_COOKIE[ 'gpollsurveyresearchtest_form_' . $form_id ] );
		$override  = false;
		if ( rgpost( 'override' ) == 1 ) {
			$show_results_link = rgpost( 'showResultsLink' ) == '1' ? true : false;

			$display_results = rgpost( 'displayResults' ) == '1' ? true : false;
			$confirmation    = rgpost( 'confirmation' ) == '1' ? true : false;
			$percentages     = rgpost( 'percentages' ) == '1' ? true : false;
			$counts          = rgpost( 'counts' ) == '1' ? true : false;
			$cookie_duration = urldecode( rgpost( 'cookieDuration' ) );
			$style           = rgpost( 'style' );
			$checksum        = rgpost( 'checksum' );
			if ( $checksum == $this->generate_checksum( $display_results, $show_results_link, $cookie_duration, $confirmation, $percentages, $counts, $style ) ) {
				$override = true;
			}
		}

		if ( false === $override ) {
			$show_results_link   = $this->get_form_setting( $form, 'showResultsLink' );
			$display_results     = $this->get_form_setting( $form, 'displayResults' );
			$confirmation        = true;
			$percentages         = $this->get_form_setting( $form, 'showPercentages' );
			$counts              = $this->get_form_setting( $form, 'showCounts' );
			$style               = $this->get_form_setting( $form, 'style' );
			$block_repeat_voters = $this->get_form_setting( $form, 'blockRepeatVoters' );

			if ( $block_repeat_voters ) {
				$cookie_duration = $this->get_form_setting( $form, 'cookie' );
			} else {
				$cookie_duration = '';
			}
		}


		$can_vote          = ( ! $has_voted ) || ( empty( $cookie_duration ) && $has_voted );
		$output['canVote'] = $can_vote;

		if ( $preview_results || ( false === $can_vote ) ) {

			if ( '' === $show_results_link ) {
				$show_results_link = true;
			}

			if ( ( $preview_results && $show_results_link ) || $display_results ) {
				$results             = $this->gpollsurveyresearchtest_get_results( $form_id, '0', $style, $percentages, $counts );
				$results_summary     = $results['summary'];
				$output['resultsUI'] = $results_summary;
			} else {
				if ( $confirmation ) {
					require_once( GFCommon::get_base_path() . '/form_display.php' );
					$output['resultsUI'] = GFFormDisplay::handle_confirmation( $form, null );
				} else {
					$output['resultsUI'] = '';
				}
			}
		} else {
			$output['resultsUI'] = '';
		}

		echo json_encode( $output );
		die();

	}


	// # CRON JOB & RESULTS CACHE --------------------------------------------------------------------------------------

	/**
	 * Adds once weekly to the existing cron schedules.
	 *
	 * @param array $schedules An array of non-default cron schedules.
	 *
	 * @return array
	 */
	public function cron_add_custom_schedule( $schedules ) {
		$custom_schedules = $this->get_custom_cron_schedule();
		$schedules        = array_merge( $schedules, $custom_schedules );

		return $schedules;
	}


	/**
	 * Used for the second parameter of the wp_schedule_event function. How often the event should recur.
	 *
	 * @return string
	 */
	public function get_cron_recurrence() {

		$custom_schedule = $this->get_custom_cron_schedule();

		if ( empty( $custom_schedule ) ) {
			$recurrence = 'hourly';
		} else {
			$recurrence = current( array_keys( $custom_schedule ) );
		}

		return $recurrence;
	}


	/**
	 * Helper to allow the gform_polls_cron_schedule filter to be used to override the default cron schedule.
	 *
	 * @return array
	 */
	public function get_custom_cron_schedule() {

		/**
		 * Allows modification to the cron schedule for Polls
		 *
		 * @param array $schedule An array to allow modifications to the cron schedule or create a custom cron schedule
		 */
		$schedule = apply_filters( 'gform_polls_cron_schedule', array() );

		return $schedule;
	}

	/**
	 * Unschedule the cron when the plugin is deactivated.
	 */
	public static function remove_wp_cron_task() {
		wp_clear_scheduled_hook( 'gform_polls_cron' );
	}

	/**
	 * Called only by the wp_cron task.
	 */
	public function wp_cron_task() {
		$forms = GFFormsModel::get_forms( true );
		foreach ( $forms as $form ) {
			$form_id     = $form->id;
			$form_meta   = $this->get_form_meta( $form_id );
			$pollsurveyresearchtest_fields = GFAPI::get_fields_by_type( $form_meta, array( 'pollsurveyresearchtest' ) );
			if ( empty ( $pollsurveyresearchtest_fields ) )
				continue;

			$data_tmp = GFCache::get( 'gpollsurveyresearchtest_data_tmp_' . $form_id );
			if ( false === $data_tmp ) {
				$data = GFCache::get( 'gpollsurveyresearchtest_data_' . $form_id );
				if ( false == $data || rgar( $data, 'incomplete' ) || false === isset( $data['execution_time'] ) || rgar( $data, 'expired' ) ) {
					$data = $this->gpollsurveyresearchtest_get_data( $form_id );
					$this->maybe_continue_cache_rebuild( $data, $form_id );
				}
			} else {
				$data = $this->gpollsurveyresearchtest_get_data( $form_id, $data_tmp );
				$this->maybe_continue_cache_rebuild( $data, $form_id );
			}
		}
	}

	/**
	 * Called only by the wp_cron job.
	 *
	 * @param array $data
	 * @param int $form_id The form ID.
	 */
	public function maybe_continue_cache_rebuild( $data, $form_id ) {
		if ( rgar( $data, 'incomplete' ) ) {
			GFCache::set( 'gpollsurveyresearchtest_data_tmp_' . $form_id, $data, true );
		} else {
			GFCache::set( 'gpollsurveyresearchtest_data_' . $form_id, $data, true );
			GFCache::delete( 'gpollsurveyresearchtest_data_tmp_' . $form_id );
		}
	}

	/**
	 * Called on entry created, entry updated, entry status changed and form saved.
	 * Not called by the wp_cron job.
	 *
	 * @param int $form_id The form ID.
	 */
	public function maybe_update_cache( $form_id ) {
		$key  = 'gpollsurveyresearchtest_data_' . $form_id;
		$data = GFCache::get( $key );
		if ( false === $data ) {
			// nothing in the cache so start building it
			$this->update_cache( $form_id );
		} else {
			if ( rgar( $data, 'execution_time' ) < 5 ) {
				// update the cache now if the last execution was under 5 seconds
				$this->update_cache( $form_id );
			} else {
				// mark the cache expired so the wp_cron job will begin recalculation
				$data['expired'] = true;
				GFCache::set( $key, $data, true );
			}
		}
	}


	/**
	 * Update the results cache when an entry is created.
	 *
	 * @param array $entry The entry which was created.
	 * @param array $form The current form.
	 */
	public function entry_created( $entry, $form ) {
		$pollsurveyresearchtest_fields = GFAPI::get_fields_by_type( $form, array( 'pollsurveyresearchtest' ) );
		if ( empty ( $pollsurveyresearchtest_fields ) ) {
			return;
		}

		//update cache
		$form_id = $form['id'];
		$this->maybe_update_cache( $form_id );

	}

	/**
	 * Update the results cache when a form is saved.
	 *
	 * @param array $form The current form.
	 * @param bool $is_new True if this is a new form being created. False if this is an existing form being updated.
	 */
	public function after_save_form( $form, $is_new ) {

		$pollsurveyresearchtest_fields = GFAPI::get_fields_by_type( $form, array( 'pollsurveyresearchtest' ) );
		if ( empty ( $pollsurveyresearchtest_fields ) ) {
			return;
		}
		//update cache
		$form_id = $form['id'];

		$this->maybe_update_cache( $form_id );
	}


	/**
	 * Cache the form meta.
	 *
	 * @param int $form_id The form ID.
	 *
	 * @return mixed
	 */
	public function get_form_meta( $form_id ) {
		$form_metas = $this->_form_meta_by_id;

		if ( empty( $form_metas ) ) {
			$form_ids = array();
			$forms    = RGFormsModel::get_forms();
			foreach ( $forms as $form ) {
				$form_ids[] = $form->id;
			}

			if ( method_exists( 'GFFormsModel', 'get_form_meta_by_id' ) )
				$form_metas = GFFormsModel::get_form_meta_by_id( $form_ids );
			else
				$form_metas = GFFormsModel::get_forms_by_id( $form_ids ); //backwards compatiblity with <1.7

			$this->_form_meta_by_id = $form_metas;
		}
		foreach ( $form_metas as $form_meta ) {
			if ( $form_meta['id'] == $form_id )
				return $form_meta;
		}

	}


	/**
	 * Update the results cache.
	 *
	 * @param int $form_id The form ID.
	 *
	 * @return array
	 */
	public function update_cache( $form_id ) {

		$gpollsurveyresearchtest_data = $this->gpollsurveyresearchtest_get_data( $form_id );
		GFCache::set( 'gpollsurveyresearchtest_data_' . $form_id, $gpollsurveyresearchtest_data, true );

		return $gpollsurveyresearchtest_data;
	}

	/**
	 * Cycles through all entries, counts responses and returns an associative array with the data for each field.
	 * It's then optionally cached later according to the user settings.
	 *
	 * @param int $form_id The form ID.
	 * @param array $gpollsurveyresearchtest_data The poll results data.
	 *
	 * @return array
	 */
	public function gpollsurveyresearchtest_get_data( $form_id, $gpollsurveyresearchtest_data = array() ) {
		$time_start         = microtime( true );
		$max_execution_time = 20; //seconds
		$totals             = RGFormsModel::get_form_counts( $form_id );
		$total              = $totals['total'];

		$form_meta   = RGFormsModel::get_form_meta( $form_id );
		$form_meta   = apply_filters( "gform_polls_form_pre_results_$form_id", apply_filters( 'gform_polls_form_pre_results', $form_meta ) );
		$pollsurveyresearchtest_fields = array();

		foreach ( $form_meta['fields'] as $field ) {
			if ( $field->type !== 'pollsurveyresearchtest' ) {
				continue;
			}
			$pollsurveyresearchtest_fields[] = clone $field;
		}


		$offset        = 0;
		$page_size     = 200;
		$field_counter = 0;

		$search_criteria = array(
			'status' => 'active',
		);

		$sorting = array(
			'key' => 0,
			'direction' => 'DESC',
			'is_numeric' => false,
		);
		
		if ( empty( $gpollsurveyresearchtest_data ) ) {

			//first build list of fields to count and later count the entries
			//it's split up this way to avoid a timeout on large resultsets

			foreach ( $pollsurveyresearchtest_fields as $pollsurveyresearchtest_field ) {

				$fieldid = $pollsurveyresearchtest_field->id;

				$gpollsurveyresearchtest_field_data = array(
					'field_label' => $pollsurveyresearchtest_field->label,
					'field_id'    => $fieldid,
					'type'        => $pollsurveyresearchtest_field->type,
					'inputType'   => $pollsurveyresearchtest_field->inputType,
				);

				$gpollsurveyresearchtest_data['fields'][ $field_counter ] = $gpollsurveyresearchtest_field_data;

				$gpollsurveyresearchtest_input_data = array();

				//for checkboxes
				if ( $pollsurveyresearchtest_field->inputType == 'checkbox' ) {
					$input_counter = 0;
					foreach ( $pollsurveyresearchtest_field->inputs as $input ) {
						$inputid = str_replace( '.', '_', $input['id'] );

						$gpollsurveyresearchtest_input_data = array(
							'input_id' => "#choice_{$inputid}",
							'label'    => $input['label'],
						);
						$gpollsurveyresearchtest_data['fields'][ $field_counter ]['inputs'][ $input_counter ] = $gpollsurveyresearchtest_input_data;
						$input_counter += 1;
					}
				} else if ( $pollsurveyresearchtest_field->inputType == 'number' ) {
					//for number
// @TODO
					$pollsurveyresearchtest_field->choices               = [];
				} else {
					//for radio & dropdowns

					$choice_counter = 0;
					if ( $pollsurveyresearchtest_field->enableOtherChoice ) {
						$choice_index                      = count( $pollsurveyresearchtest_field->choices );
						$choices                           = $pollsurveyresearchtest_field->choices;
						$choices[ $choice_index ]['text']  = esc_html__( 'Other', 'brewpixel-gravityforms-polls-surveys-researchtests' );
						$choices[ $choice_index ]['value'] = 'gpollsurveyresearchtest_other';
						$pollsurveyresearchtest_field->choices               = $choices;
					}

					foreach ( $pollsurveyresearchtest_field->choices as $choice ) {
						$gpollsurveyresearchtest_input_data                                                = array(
							'input_id' => "#choice_{$fieldid}_{$choice_counter}",
							'label'    => $choice['text'],
						);
						$gpollsurveyresearchtest_data['fields'][ $field_counter ]['inputs'][ $choice_counter ] = $gpollsurveyresearchtest_input_data;
						$choice_counter += 1;
					}
				}
				$field_counter += 1;
				$i = $offset;

			}
		} else {
			$i = $gpollsurveyresearchtest_data['offset'];
			unset( $gpollsurveyresearchtest_data['offset'] );
		}

		$precision = apply_filters( 'gform_polls_percentage_precision', 0, $form_id );

		//get leads in groups of $page_size to avoid timeouts
		while ( $i <= $total ) {
			$field_counter = 0;

			$get_leads_time_start = microtime( true );

			$paging = array(
				'offset' => $i,
				'page_size' => $page_size,
			);

			$entries = GFAPI::get_entries( $form_id, $search_criteria, $sorting, $paging );

			$get_leads_time_end   = microtime( true );
			$get_leads_time       = $get_leads_time_end - $get_leads_time_start;

			//loop through each field currently on the form and count the entries for each choice
			foreach ( $pollsurveyresearchtest_fields as $pollsurveyresearchtest_field ) {

				if ( isset ( $gpollsurveyresearchtest_data['fields'][ $field_counter ]['total_entries'] ) ) {
					$field_total_entries = $gpollsurveyresearchtest_data['fields'][ $field_counter ]['total_entries'];
				} else {
					$field_total_entries = 0;
				}

				foreach ( $entries as $entry ) {
					$entry_value = RGFormsModel::get_lead_field_value( $entry, $pollsurveyresearchtest_field );

					if ( false === empty( $entry_value ) )
						$field_total_entries ++;
				}
				$gpollsurveyresearchtest_data['fields'][ $field_counter ]['total_entries'] = $field_total_entries;

				$gpollsurveyresearchtest_input_data = array();

				// checkboxes store entries differently to radio & dropdowns
				if ( $pollsurveyresearchtest_field->inputType == 'checkbox' ) {
					//for checkboxes

					// loop through all the choices and count the entries for each choice
					$input_counter = 0;
					foreach ( $pollsurveyresearchtest_field->inputs as $input ) {

						// running total of entries for each set of entries
						if ( isset ( $gpollsurveyresearchtest_data['fields'][ $field_counter ]['inputs'][ $input_counter ]['total_entries'] ) ) {
							$total_entries = $gpollsurveyresearchtest_data['fields'][ $field_counter ]['inputs'][ $input_counter ]['total_entries'];
						} else {
							$total_entries = 0;
						}
						$entry_index = 1;

						// loop through all the entries and count the entries for the choice
						foreach ( $entries as $entry ) {

							// loop through each item in the lead object and pick out the entries for this field id
							foreach ( $entry as $key => $entry_value ) {

								// checkboxes store the key as [field number].[input index] (e.g. 2.1 or 2.2)
								// so convert to integer to identify all the responses inside the lead object for this field id
								if ( intval( $key ) == $pollsurveyresearchtest_field->id ) {
									//compare the user's response with the current choice
									if ( $entry_value == $pollsurveyresearchtest_field->choices[ $input_counter ]['value'] ) {
										// found a response for this choice so continue to the next lead
										$total_entries ++;
										break;
									}
								}
							}
							$entry_index += 1;
						}

						//calculate the ratio of total number of responses counted to the total number of entries for this form
						$ratio = 0;

						if ( $field_total_entries != 0 ) {
							$ratio = round( ( $total_entries / $field_total_entries * 100 ), $precision );
						}

						//store the data
						$gpollsurveyresearchtest_data['fields'][ $field_counter ]['inputs'][ $input_counter ]['value']         = $pollsurveyresearchtest_field->choices[ $input_counter ]['value'];
						$gpollsurveyresearchtest_data['fields'][ $field_counter ]['inputs'][ $input_counter ]['total_entries'] = $total_entries;
						$gpollsurveyresearchtest_data['fields'][ $field_counter ]['inputs'][ $input_counter ]['ratio']         = $ratio;
						$input_counter += 1;
					}
				// number store entries differently to checkboxes, radio & dropdowns
				} else if ( $pollsurveyresearchtest_field->inputType == 'number' ) {
					//for numbers
					$pollsurveyresearchtest_fieldinputs = array_keys($this->array_unique_deep($entries, $pollsurveyresearchtest_field->id));
					
					// foreach ( $entries as $entry ) {
					// 	foreach ( $entry as $key => $entry_value ) {
					// 		if ( intval( $key ) == 4 ) {
					// 		echo'key ';
					// 		echo $key;
					// 		echo "\tvalue ";
					// 		echo $entry_value;
					// 		echo '<br>';
					// 		echo '<br>';
					// 	}
					// 	}
					// }
					// die();
					// loop through all the choices and count the entries for each choice
					$input_counter = 0;

					// @TODO Find all the unique entries and make those equal to the $pollsurveyresearchtest_field->inputs
					foreach ( $pollsurveyresearchtest_fieldinputs as $input ) {

						// running total of entries for each set of entries
						if ( isset ( $gpollsurveyresearchtest_data['fields'][ $field_counter ]['inputs'][ $input_counter ]['total_entries'] ) ) {
							$total_entries = $gpollsurveyresearchtest_data['fields'][ $field_counter ]['inputs'][ $input_counter ]['total_entries'];
						} else {
							$total_entries = 0;
						}
						$entry_index = 1;

						// loop through all the entries and count the entries for the choice
						foreach ( $entries as $entry ) {

							// loop through each item in the lead object and pick out the entries for this field id
							foreach ( $entry as $key => $entry_value ) {

								// checkboxes store the key as [field number].[input index] (e.g. 2.1 or 2.2)
								// so convert to integer to identify all the responses inside the lead object for this field id
								if ( intval( $key ) == $pollsurveyresearchtest_field->id ) {
									//compare the user's response with the current choice
									if ( $entry_value == $input ) {
										// found a response for this choice so continue to the next lead
										$total_entries ++;
										break;
									}
								}
							}
							$entry_index += 1;
						}

						//calculate the ratio of total number of responses counted to the total number of entries for this form
						$ratio = 0;

						if ( $field_total_entries != 0 ) {
							$ratio = round( ( $total_entries / $field_total_entries * 100 ), $precision );
						}

						//store the data
						$gpollsurveyresearchtest_data['fields'][ $field_counter ]['inputs'][ $input_counter ]['value']         = $input;
						$gpollsurveyresearchtest_data['fields'][ $field_counter ]['inputs'][ $input_counter ]['total_entries'] = $total_entries;
						$gpollsurveyresearchtest_data['fields'][ $field_counter ]['inputs'][ $input_counter ]['ratio']         = $ratio;
						$input_counter += 1;
					}
				} else {
					// for radio & dropdowns

					$choice_counter = 0;

					// loop through each choice and count the responses
					foreach ( $pollsurveyresearchtest_field->choices as $choice ) {

						// running total of entries for each set of entries
						if ( isset ( $gpollsurveyresearchtest_data['fields'][ $field_counter ]['inputs'][ $choice_counter ]['total_entries'] ) ) {
							$total_entries = $gpollsurveyresearchtest_data['fields'][ $field_counter ]['inputs'][ $choice_counter ]['total_entries'];
						} else {
							$total_entries = 0;
						}

						// count responses for 'Other'
						if ( rgar( $choice, 'value') == 'gpollsurveyresearchtest_other' ) {
							$possible_choices = array();
							foreach ( $pollsurveyresearchtest_field->choices as $possible_choice ) {
								array_push( $possible_choices, rgar( $possible_choice, 'value') );
							}

							foreach ( $entries as $entry ) {
								$entry_value = RGFormsModel::get_lead_field_value( $entry, $pollsurveyresearchtest_field );

								if ( ! empty( $entry_value ) && ! in_array( $entry_value, $possible_choices ) ) {
									$total_entries ++;
								}
							}
						} else {

							// count entries
							foreach ( $entries as $entry ) {
								$entry_value = RGFormsModel::get_lead_field_value( $entry, $pollsurveyresearchtest_field );
								if ( $entry_value === rgar( $choice, 'value' ) ) {
									$total_entries ++;
								}

							}
						}

						// calculate the ratio of total number of responses counted to the total number of entries for this form
						$ratio = 0;
						if ( $field_total_entries != 0 ) {
							$ratio = round( ( $total_entries / $field_total_entries * 100 ), $precision );
						}


						//store the data
						$gpollsurveyresearchtest_data['fields'][ $field_counter ]['inputs'][ $choice_counter ]['value']         = rgar( $choice, 'value' );
						$gpollsurveyresearchtest_data['fields'][ $field_counter ]['inputs'][ $choice_counter ]['total_entries'] = $total_entries;
						$gpollsurveyresearchtest_data['fields'][ $field_counter ]['inputs'][ $choice_counter ]['ratio']         = $ratio;
						$choice_counter += 1;
					}
				}

				$field_counter += 1;
			}
			$i += $page_size;
			$time_end                     = microtime( true );
			$execution_time               = ( $time_end - $time_start );
			$gpollsurveyresearchtest_data['execution_time'] = isset( $gpollsurveyresearchtest_data['execution_time'] ) ? $gpollsurveyresearchtest_data['execution_time'] + $execution_time : $execution_time;
			$gpollsurveyresearchtest_data['incomplete']     = false;
			if ( $execution_time + $get_leads_time > $max_execution_time ) {
				$gpollsurveyresearchtest_data['incomplete'] = true;
				$gpollsurveyresearchtest_data['offset']     = $i;
				break;
			}
		} //end while
		return $gpollsurveyresearchtest_data;
	} // end function gpollsurveyresearchtest_get_data

	/**
	* Returns an array where the keys are the unique values and the values are the number of occurences in the original array.
	*/
	public function array_unique_deep($array, $key) {
		$values = array();
		foreach ($array as $k1 => $row) {
	    	foreach ($row as $k2 => $v) {
	      		if ($k2 == $key) {
	        		$values[ $k1 ] = $v;
	        		continue;
	      		}
	    	}
	  	}
	  	return array_count_values($values);
	}

	/**
	 * Returns the results in an array of HTML formatted data.
	 *
	 * @param int $formid The form ID.
	 * @param string $display_field
	 * @param string $style
	 * @param bool $show_percentages
	 * @param bool $show_counts
	 * @param array $lead
	 *
	 * @return array
	 */
	public function gpollsurveyresearchtest_get_results( $formid, $display_field = '0' /* zero = all fields */, $style = 'green', $show_percentages = true, $show_counts = true, $lead = array() ) {

		$gpollsurveyresearchtest_output = array();
		$gpollsurveyresearchtest_data   = array();

		// each bar will receive this HTML formatting
		$bar_html = "<div class='gpollsurveyresearchtest_wrapper {$style}'><div class='gpollsurveyresearchtest_ratio_box'><div class='gpollsurveyresearchtest_ratio_label'>%s</div></div><div class='gpollsurveyresearchtest_bar'>";
		$bar_html .= "<span class='gpollsurveyresearchtest_bar_juice' data-origwidth='%s' style='width: %s%%'><span class='gpollsurveyresearchtest_bar_count'>%s</span></span></div></div>";


		// if data is cached then pull the data out of the cache

		if ( false === ( $gpollsurveyresearchtest_data = GFCache::get( 'gpollsurveyresearchtest_data_' . $formid ) ) ) {

			// cache has timed out so get the data again and cache it again
			$gpollsurveyresearchtest_data = $this->update_cache( $formid );
		}


		// build HTML output

		$gpollsurveyresearchtest_output['summary'] = "<div class='gpollsurveyresearchtest_container'>";
		$field_counter           = 0;

		// loop through polls data field by field
		foreach ( $gpollsurveyresearchtest_data['fields'] as $field ) {
			
			if ( $field['inputType'] === 'number' ) {
				$fieldid = $field['field_id'];

				// only build html for the field(s) specified in the parameter. 0 = all fields
				if ( is_array( $display_field ) ) {
					if ( false === in_array( $fieldid, $display_field ) )
						continue;
				} elseif ( $display_field != '0' && $fieldid != $display_field ) {
					continue;
				}


				// build 2 sections: summary and individual fields
				$field_number = $field_counter + 1;
				$gpollsurveyresearchtest_output['summary'] .= "<div class='gpollsurveyresearchtest_field'>";
				$gpollsurveyresearchtest_output['summary'] .= "<div class='gpollsurveyresearchtest_field_label_container'>";

				$gpollsurveyresearchtest_output['summary'] .= "<div class='gpollsurveyresearchtest_field_label'>";
				$gpollsurveyresearchtest_output['summary'] .= $field['field_label'] . '</div></div>';

				// the individual fields HTML was used in the past but not used now.
				// it was used to display results 'inline' with the form (i.e. form input then the bar below)
				// I've left it because it may be useful either to designers or for a future use
				$gpollsurveyresearchtest_output['fields'][ $field_counter ]['field_id'] = $field['field_id'];
				$gpollsurveyresearchtest_output['fields'][ $field_counter ]['type']     = $field['type'];

				$selected_values = array();

				// if the lead object is passed then prepare to highlight the selected choices
				if ( ! empty ( $lead ) ) {
					$form_meta = RGFormsModel::get_form_meta( $formid );
					// collect all the responses in the lead for this field
					$selected_values = $this->get_selected_values( $form_meta, $fieldid, $lead );

					//collect all the choices that are currently possible in the field

					$possible_choices = $this->get_possible_choices( $form_meta, $fieldid );

					$form_meta_field = RGFormsModel::get_field( $form_meta, $fieldid );

					// if the 'other' option is selected for this field
					// add the psuedo-value 'gpollsurveyresearchtest_other' if responses are found that are not in the list of possible choices
					if ( $form_meta_field->enableOtherChoice ) {

						foreach ( $selected_values as $selected_value ) {
							if ( ! in_array( $selected_value, $possible_choices ) )
								array_push( $selected_values, 'gpollsurveyresearchtest_other' );
						}
					}
				}

				// loop through all the inputs in this field (poll data field not form object field) and build the HTML for the bar
				$input_counter = 0;
				foreach ( $field['inputs'] as $input ) {
					if ( '' !== rgar( $input, 'value' ) ) {
						//highlight the selected value by adding a class to the label
						$selected_class = '';
						if ( in_array( rgar( $input, 'value' ), $selected_values ) ) {
							$selected_class .= ' gpollsurveyresearchtest_value_selected';
						}

						//build the bar and add it to the summary
						$gpollsurveyresearchtest_output['summary'] .= sprintf( "<div class='gpollsurveyresearchtest_choice_label%s'>%s</div>", $selected_class, rgar( $input, 'value' ) );
						$ratio            = rgar( $input, 'ratio' );
						$count            = $show_counts === true ? rgar( $input, 'total_entries' ) : '';
						$percentage_label = $show_percentages === true ? $ratio . '%' : '';
						$input_html       = sprintf( $bar_html, $percentage_label, $ratio, $ratio, $count );
						$gpollsurveyresearchtest_output['summary'] .= $input_html;
						//print_r($input);
						//add the bar HTML to the fields array ready to output alongside the summary
						$input_data                                                       = array(
							'input_id'      => rgar( $input, 'input_id' ),
							'label'         => rgar( $input, 'value' ),
							'total_entries' => $input['total_entries'],
							'ratio'         => $input['ratio'],
							'bar_html'      => $input_html,
						);

						$gpollsurveyresearchtest_output['fields'][ $field_counter ]['inputs'][ $input_counter ] = $input_data;

						$input_counter += 1;
					}
				}
				$gpollsurveyresearchtest_output['summary'] .= '</div>';
				$field_counter += 1;
			} else {
				$fieldid = $field['field_id'];

				// only build html for the field(s) specified in the parameter. 0 = all fields
				if ( is_array( $display_field ) ) {
					if ( false === in_array( $fieldid, $display_field ) )
						continue;
				} elseif ( $display_field != '0' && $fieldid != $display_field ) {
					continue;
				}


				// build 2 sections: summary and individual fields
				$field_number = $field_counter + 1;
				$gpollsurveyresearchtest_output['summary'] .= "<div class='gpollsurveyresearchtest_field'>";
				$gpollsurveyresearchtest_output['summary'] .= "<div class='gpollsurveyresearchtest_field_label_container'>";

				$gpollsurveyresearchtest_output['summary'] .= "<div class='gpollsurveyresearchtest_field_label'>";
				$gpollsurveyresearchtest_output['summary'] .= $field['field_label'] . '</div></div>';

				// the individual fields HTML was used in the past but not used now.
				// it was used to display results 'inline' with the form (i.e. form input then the bar below)
				// I've left it because it may be useful either to designers or for a future use
				$gpollsurveyresearchtest_output['fields'][ $field_counter ]['field_id'] = $field['field_id'];
				$gpollsurveyresearchtest_output['fields'][ $field_counter ]['type']     = $field['type'];

				$selected_values = array();

				// if the lead object is passed then prepare to highlight the selected choices
				if ( ! empty ( $lead ) ) {
					$form_meta = RGFormsModel::get_form_meta( $formid );
					// collect all the responses in the lead for this field
					$selected_values = $this->get_selected_values( $form_meta, $fieldid, $lead );

					//collect all the choices that are currently possible in the field

					$possible_choices = $this->get_possible_choices( $form_meta, $fieldid );

					$form_meta_field = RGFormsModel::get_field( $form_meta, $fieldid );

					// if the 'other' option is selected for this field
					// add the psuedo-value 'gpollsurveyresearchtest_other' if responses are found that are not in the list of possible choices
					if ( $form_meta_field->enableOtherChoice ) {

						foreach ( $selected_values as $selected_value ) {
							if ( ! in_array( $selected_value, $possible_choices ) )
								array_push( $selected_values, 'gpollsurveyresearchtest_other' );
						}
					}
				}

				// loop through all the inputs in this field (poll data field not form object field) and build the HTML for the bar
				$input_counter = 0;
				foreach ( $field['inputs'] as $input ) {

					//highlight the selected value by adding a class to the label
					$selected_class = '';
					if ( in_array( rgar( $input, 'value' ), $selected_values ) ) {
						$selected_class .= ' gpollsurveyresearchtest_value_selected';
					}

					//build the bar and add it to the summary
					$gpollsurveyresearchtest_output['summary'] .= sprintf( "<div class='gpollsurveyresearchtest_choice_label%s'>%s</div>", $selected_class, rgar( $input, 'label' ) );
					$ratio            = rgar( $input, 'ratio' );
					$count            = $show_counts === true ? rgar( $input, 'total_entries' ) : '';
					$percentage_label = $show_percentages === true ? $ratio . '%' : '';
					$input_html       = sprintf( $bar_html, $percentage_label, $ratio, $ratio, $count );
					$gpollsurveyresearchtest_output['summary'] .= $input_html;
					//print_r($input);
					//add the bar HTML to the fields array ready to output alongside the summary
					$input_data                                                       = array(
						'input_id'      => rgar( $input, 'input_id' ),
						'label'         => rgar( $input, 'label' ),
						'total_entries' => $input['total_entries'],
						'ratio'         => $input['ratio'],
						'bar_html'      => $input_html,
					);

					$gpollsurveyresearchtest_output['fields'][ $field_counter ]['inputs'][ $input_counter ] = $input_data;

					$input_counter += 1;
				}
				$gpollsurveyresearchtest_output['summary'] .= '</div>';
				$field_counter += 1;
			}
		}
		$gpollsurveyresearchtest_output['summary'] .= '</div>';

		return $gpollsurveyresearchtest_output;

	} //end function gpollsurveyresearchtest_get_results

	/**
	 * Collect all the responses in the lead for this field and returns an array.
	 *
	 * @param array $form_meta The current form meta.
	 * @param int $fieldid The current field ID.
	 * @param array $lead The current entry.
	 *
	 * @return array
	 */
	public function get_selected_values( $form_meta, $fieldid, $lead ) {

		$selected_values = array();

		//pick out the field we need from the fields collection in the form object
		//and add the selected values to the selected_values array
		if ( is_array( $form_meta['fields'] ) ) {
			foreach ( $form_meta['fields'] as $field ) {
				if ( $field->id == $fieldid ) {
					if ( $field->inputType == 'checkbox' ) {
						for ( $i = 1; $i <= count( $field->inputs ); $i ++ ) {
							$lead_index = 0;
							$lead_index = $fieldid . '.' . $i;
							if ( isset( $lead[ $lead_index ] ) && ! empty( $lead[ $lead_index ] ) )
								array_push( $selected_values, $lead[ $lead_index ] );
						}
					} else {
						for ( $i = 1; $i <= count( $field->choices ); $i ++ ) {
							$lead_index = $fieldid;
							if ( isset( $lead[ $lead_index ] ) && ! empty( $lead[ $lead_index ] ) )
								array_push( $selected_values, $lead[ $lead_index ] );
						}
					}
					break;
				}
			}
		}

		return $selected_values;
	}

	/**
	 * Get the choice values for the specified field.
	 *
	 * @param array $form_meta The current form meta.
	 * @param int $fieldid The current field ID.
	 *
	 * @return array
	 */
	public function get_possible_choices( $form_meta, $fieldid ) {

		$possible_choices = array();

		//pick out the field we need from the fields collection in the form object
		//and add the possible choices to the possible_choices array
		if ( is_array( $form_meta['fields'] ) ) {
			foreach ( $form_meta['fields'] as $field ) {
				if ( $field->id == $fieldid && ! empty( $field->choices ) ) {
					foreach ( $field->choices as $possible_choice ) {
						array_push( $possible_choices, rgar( $possible_choice, 'value' ) );
					}

					return $possible_choices;
				}
			}
		}
	}


	// # SHORTCODES ----------------------------------------------------------------------------------------------------

	/**
	 * Displays the form and specifies hidden form values to enable and configure the poll.
	 * If the cookie is already set then display the results.
	 *
	 * @param $string
	 * @param $attributes
	 * @param $content
	 *
	 * @return mixed|null|string|void
	 */
	function pollsurveyresearchtest_shortcode( $string, $attributes, $content ) {

		extract(
			shortcode_atts(
				array(
					'title'             => true,
					'description'       => true,
					'confirmation'      => false,
					'id'                => 0,
					'name'              => '',
					'field_values'      => '',
					'ajax'              => false,
					'disable_scripts'   => false,
					'tabindex'          => 1,
					'mode'              => 'pollsurveyresearchtest',
					'field'             => 0,
					'style'             => 'green',
					'display_results'   => true,
					'show_results_link' => true,
					'percentages'       => true,
					'counts'            => true,
					'cookie'            => '',

				), $attributes
			)
		);


		$currentDate            = strtotime( 'now' );
		$cookie                 = strtolower( $cookie );
		$cookie_expiration_date = date( strtotime( $cookie ) );

		$confirmation = strtolower( $confirmation ) == 'false' ? false : true;
		if ( ! empty( $cookie ) && $cookie_expiration_date <= $currentDate ) {
			return sprintf( esc_html__( 'Gravity Forms Polls Add-on Shortcode error: Please enter a valid date or time period for the cookie expiration cookie_expiration_date: %s', 'brewpixel-gravityforms-polls-surveys-researchtests' ), $cookie_expiration_date );
		}

		$percentages       = strtolower( $percentages ) == 'false' ? false : true;
		$counts            = strtolower( $counts ) == 'false' ? false : true;
		$display_results   = strtolower( $display_results ) == 'false' ? false : true;
		$show_results_link = strtolower( $show_results_link ) == 'false' ? false : true;

		$title           = strtolower( $title ) == 'false' ? false : true;
		$description     = strtolower( $description ) == 'false' ? false : true;
		$ajax            = strtolower( $ajax ) == 'true' ? true : false;
		$disable_scripts = strtolower( $disable_scripts ) == 'true' ? true : false;

		$return = true;

		$poll_ui = $this->build_poll_ui( $id, $field, $style, $mode, $percentages, $counts, $title, $description, $confirmation, $show_results_link, $ajax, $cookie, $display_results, $field_values, $disable_scripts, $tabindex, $return );

		return $poll_ui;

	} // end function pollsurveyresearchtest_shortcode

	public function build_poll_ui( $form_id, $field_id = 0 /* zero = all fields */, $style = 'green', $mode = 'pollsurveyresearchtest', $percentages = true, $counts = true, $title = true, $description = true, $confirmation = false, $show_results_link, $ajax = false, $cookie = '', $display_results = true, $field_values = '', $disable_scripts = false, $tabindex, $return = true ) {
		$form = RGFormsModel::get_form_meta( $form_id );
		if ( empty( $form ) ) {
			return;
		}

		$pollsurveyresearchtest_fields = GFAPI::get_fields_by_type( $form, array( 'pollsurveyresearchtest' ) );

		if ( empty( $pollsurveyresearchtest_fields ) ) {
			return;
		}

		$this->gpollsurveyresearchtest_add_scripts = true;

		if ( $mode == 'results' ) {

			$results = $this->gpollsurveyresearchtest_get_results( $form_id, $field_id, $style, $percentages, $counts );
			$output  = $results['summary'];


		} else {
			$checksum          = $this->generate_checksum( $display_results, $show_results_link, $cookie, $confirmation, $percentages, $counts, $style );
			$show_results_link = false === $show_results_link ? 0 : 1;

			$field_values = htmlspecialchars_decode( $field_values );
			$field_values = str_replace( '&#038;', '&', $field_values );

			$percentages     = $percentages === false ? 0 : 1;
			$counts          = $counts === false ? 0 : 1;
			$display_results = $display_results ? 1 : 0;

			if ( $disable_scripts === false ) {
				RGForms::print_form_scripts( $form, $ajax );
			}


			if ( $field_values != '' ) {
				$field_values .= '&';
			}
			$field_values .= "gpollsurveyresearchtest_enabled=1&gpollsurveyresearchtest_field={$field_id}&gpollsurveyresearchtest_style={$style}&gpollsurveyresearchtest_display_results={$display_results}&gpollsurveyresearchtest_show_results_link={$show_results_link}&gpollsurveyresearchtest_cookie={$cookie}&gpollsurveyresearchtest_confirmation={$confirmation}&gpollsurveyresearchtest_percentages={$percentages}&gpollsurveyresearchtest_counts={$counts}&gpollsurveyresearchtest_checksum={$checksum}";

			parse_str( $field_values, $field_value_array );
			$field_value_array = stripslashes_deep( $field_value_array );

			$output = RGForms::get_form( $form_id, $title, $description, false, $field_value_array, $ajax, $tabindex );


		}

		if ( false === $return ) {
			echo $output;
		} else {
			return $output;
		}
	}

	private function generate_checksum( $display_results, $show_results_link, $cookie, $confirmation, $percentages, $counts, $style ) {
		$checksum_vars = array( $display_results, $show_results_link, $cookie, $confirmation, $percentages, $counts, $style );

		return wp_hash( crc32( base64_encode( serialize( $checksum_vars ) ) ) );

	}

	public function pollsurveyresearchtest_total_shortcode( $atts, $content = null ) {

		extract(
			shortcode_atts(
				array(
					'id' => '1',
				), $atts
			)
		);

		$totals = RGFormsModel::get_form_counts( $id );
		$total  = $totals['total'];

		return $total;
	}

	public function add_pollssurveysresearchtests_shortcode_ui_action( $actions ) {
		$actions[] = array(
			'pollssurveysresearchtests' => array(
				'label' => esc_html__( 'Polls', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
				'attrs' => array(
					array(
						'label'   => esc_html__( 'Style', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
						'attr'    => 'style',
						'type'    => 'select',
						'default' => 'green',
						'options' => array(
							'green'  => esc_html__( 'Green', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
							'red'    => esc_html__( 'Red', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
							'orange' => esc_html__( 'Orange', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
							'blue'   => esc_html__( 'Blue', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
						),
						'tooltip' => esc_html__( 'The Add-On currently supports 4 built-in styles: red, green, orange, blue. Defaults to "green".', 'brewpixel-gravityforms-polls-surveys-researchtests' )
					),
					array(
						'label'   => esc_html__( 'Mode', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
						'attr'    => 'mode',
						'type'    => 'select',
						'default' => 'pollsurveyresearchtest',
						'options' => array(
							'pollsurveyresearchtest'    => 'Poll Survey Research Test',
							'results' => 'Results',
						)
					),
					array(
						'label'   => esc_html__( 'Cookie', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
						'attr'    => 'cookie',
						'type'    => 'text',
						'tooltip' => esc_html__( 'Enables blocking of repeat voters. You enable this by passing a defined time period. Available time periods are: 1 day, 1 week, 1 month, or a specific date in the YYYY-MM-DD date format. Defaults to an empty string, which means no repeat voters are blocked.', 'brewpixel-gravityforms-polls-surveys-researchtests' )
					),
					array(
						'label'   => esc_html__( 'Results link', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
						'attr'    => 'show_results_link',
						'type'    => 'checkbox',
						'default' => 'true',
						'tooltip' => esc_html__( 'Display a link to view poll results without submitting the form? Supported values are: true, false. Defaults to "true".', 'brewpixel-gravityforms-polls-surveys-researchtests' )
					),
					array(
						'label'   => esc_html__( 'Display Results', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
						'attr'    => 'display_results',
						'type'    => 'checkbox',
						'default' => 'true',
						'tooltip' => esc_html__( 'Display poll results automatically when the form is submitted? Supported values are: true, false. Defaults to "true".', 'brewpixel-gravityforms-polls-surveys-researchtests' )
					),
					array(
						'label'   => esc_html__( 'Display Percentages', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
						'attr'    => 'percentages',
						'type'    => 'checkbox',
						'default' => 'true',
						'tooltip' => esc_html__( 'Display results percentages as part of results? Supported values are: true, false. Defaults to "true".', 'brewpixel-gravityforms-polls-surveys-researchtests' )
					),
					array(
						'label'   => esc_html__( 'Display Counts', 'brewpixel-gravityforms-polls-surveys-researchtests' ),
						'attr'    => 'counts',
						'type'    => 'checkbox',
						'default' => 'true',
						'tooltip' => esc_html__( 'Display number of times each choice has been selected when displaying results? Supported values are: true, false. Defaults to "true".', 'brewpixel-gravityforms-polls-surveys-researchtests' )
					),
				),
			)

		);

		return $actions;
	}

	// # CONTACTS INTEGRATION -------------------------------------------------------------------------------------------

	public function add_tab_to_contact_detail( $tabs, $contact_id ) {
		if ( $contact_id > 0 ) {
			$tabs[] = array( 'name' => 'polls', 'label' => __( 'Poll Entries', 'brewpixel-gravityforms-polls-surveys-researchtests' ) );
		}

		return $tabs;
	}

	public function contacts_tab( $contact_id ) {

		if ( false === empty( $contact_id ) ) :
			$search_criteria['status'] = 'active';
			$search_criteria['field_filters'][] = array( 'type'  => 'meta',
			                                             'key'   => 'gcontacts_contact_id',
			                                             'value' => $contact_id
			);
			$form_ids                           = array();
			$forms                              = GFFormsModel::get_forms( true );
			foreach ( $forms as $form ) {
				$form_meta   = GFFormsModel::get_form_meta( $form->id );
				$pollsurveyresearchtest_fields = GFAPI::get_fields_by_type( $form_meta, array( 'pollsurveyresearchtest' ) );
				if ( ! empty( $pollsurveyresearchtest_fields ) ) {
					$form_ids[] = $form->id;
				}
			}

			if ( empty( $form_ids ) ) {
				return;
			}
			$entries                   = GFAPI::get_entries( $form_ids, $search_criteria );

			if ( empty( $entries ) ) :
				esc_html_e( 'This contact has not submitted any poll entries yet.', 'brewpixel-gravityforms-polls-surveys-researchtests' );

			else : ?>

				<h3><span><?php esc_html_e( 'Poll Entries', 'brewpixel-gravityforms-polls-surveys-researchtests' ) ?></span></h3>
				<div>
					<table id="gcontacts-entry-list" class="widefat">
						<tr class="gcontacts-entries-header">
							<td>
								<?php esc_html_e( 'Entry Id', 'brewpixel-gravityforms-polls-surveys-researchtests' ) ?>
							</td>
							<td>
								<?php esc_html_e( 'Date', 'brewpixel-gravityforms-polls-surveys-researchtests' ) ?>
							</td>
							<td>
								<?php esc_html_e( 'Form', 'brewpixel-gravityforms-polls-surveys-researchtests' ) ?>
							</td>
						</tr>
						<?php
						foreach ( $entries as $entry ) {
							$form_id    = $entry['form_id'];
							$form       = GFFormsModel::get_form_meta( $form_id );
							$form_title = rgar( $form, 'title' );
							$entry_id   = $entry['id'];
							$entry_date = GFCommon::format_date( rgar( $entry, 'date_created' ), false );
							$entry_url  = admin_url( "admin.php?page=gf_entries&view=entry&id={$form_id}&lid={$entry_id}" );

							?>
							<tr>
								<td>
									<a href="<?php echo $entry_url; ?>"><?php echo $entry_id; ?></a>
								</td>
								<td>
									<?php echo $entry_date; ?>
								</td>
								<td>
									<?php echo $form_title; ?>
								</td>
							</tr>
							<?php
						}
						?>
					</table>
				</div>
				<?php
			endif;
		endif;
	}

} //end class GFPollsSurveysResearchTests
